#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Module contains functions to plot neurons in 2D/2.5D."""

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.colors as mcl
from mpl_toolkits.mplot3d.art3d import (
    Line3DCollection,
    Poly3DCollection,
    Path3DCollection,
    Patch3DCollection,
)
from matplotlib.collections import LineCollection, PolyCollection
from matplotlib.cm import ScalarMappable

import numpy as np
import pandas as pd

import pint
import warnings

from typing import Union, List, Tuple
import copy

from .. import utils, config, core, conversion
from .colors import prepare_colormap, vertex_colors
from .plot_utils import segments_to_coords, tn_pairs_to_coords
from .settings import Matplotlib2dSettings

__all__ = ["plot2d"]

logger = config.get_logger(__name__)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    pint.Quantity([])

# Default colormap for depth coloring
DEPTH_CMAP = mpl.cm.jet


def plot2d(
    x: Union[
        core.NeuronObject,
        core.Volume,
        np.ndarray,
        List[Union[core.NeuronObject, np.ndarray, core.Volume]],
    ],
    **kwargs,
) -> Tuple[mpl.figure.Figure, mpl.axes.Axes]:
    """Generate 2D plots of neurons and neuropils.

    The main advantage of this is that you can save plot as vector graphics.

    Note
    ----
    This function uses `matplotlib` which "fakes" 3D as it has only very limited
    control over layering objects in 3D. Therefore neurites are not necessarily
    plotted in the right Z order. This becomes especially troublesome when
    plotting a complex scene with lots of neurons criss-crossing. See the
    `method` parameter for details.

    Parameters
    ----------
    x :                 TreeNeuron | MeshNeuron | NeuronList | Volume | Dotprops | np.ndarray
                        Objects to plot:
                         - multiple objects can be passed as list (see examples)
                         - numpy array of shape (N, 3) is intepreted as points for
                           scatter plots

    Object parameters
    -----------------
    soma :              bool | dict, default=True

                        Plot soma if one exists. Size of the soma is determined
                        by the neuron's `.soma_radius` property which defaults
                        to the "radius" column for `TreeNeurons`. You can also
                        pass `soma` as a dictionary to customize the appearance
                        of the soma - for example `soma={"color": "red", "lw": 2, "ec": 1}`.

    radius :            bool | "auto", default=False

                        If "auto" will plot neurites of `TreeNeurons` with radius
                        if they have radii. If True, will try plotting neurites of
                        `TreeNeurons` with radius regardless. The radius can be
                        scaled by `linewidth`. Note that this will increase rendering
                        time.

    linewidth :         int | float, default=.5

                        Width of neurites. Also accepts alias `lw`.

    linestyle :         str, default='-'

                        Line style of neurites. Also accepts alias `ls`.

    color :             None | str | tuple | list | dict, default=None

                        Use single str (e.g. `'red'`) or `(r, g, b)` tuple
                        to give all neurons the same color. Use `list` of
                        colors to assign colors: `['red', (1, 0, 1), ...].
                        Use `dict` to map colors to neuron IDs:
                        `{id: (r, g, b), ...}`.

    palette :           str | array | list of arrays, default=None

                        Name of a matplotlib or seaborn palette. If `color` is
                        not specified will pick colors from this palette.

    color_by :          str | array | list of arrays, default = None

                        Color neurons by a property. Can be:
                          - a list/array of labels, one per each neuron
                          - a neuron property (str)
                          - a column name in the node table of `TreeNeurons`
                          - a list/array of values for each node
                        Numerical values will be normalized. You can control
                        the normalization by passing a `vmin` and/or `vmax` parameter.

    shade_by :          str | array | list of arrays, default=None

                        Similar to `color_by` but will affect only the alpha
                        channel of the color. If `shade_by='strahler'` will
                        compute Strahler order if not already part of the node
                        table (TreeNeurons only). Numerical values will be
                        normalized. You can control the normalization by passing
                        a `smin` and/or `smax` parameter.

    alpha :             float [0-1], default=1

                        Alpha value for neurons. Overriden if alpha is provided
                        as fourth value in `color` (rgb*a*). You can override
                        alpha value for connectors by using `cn_alpha`.

    mesh_shade :        bool, default=False

                        Only relevant for meshes (e.g. `MeshNeurons`) and
                        `TreeNeurons` with radius, and when method is 3d or
                        3d complex. Whether to shade the object which will give it
                        a 3D look.

    depth_coloring :    bool, default=False

                        If True, will use neuron color to encode depth (Z).
                        Overrides `color` argument. Does not work with
                        `method = '3d_complex'`.

    depth_scale :       bool, default=True

                        If True and `depth_coloring=True` will plot a scale.

    connectors :        bool | "presynapses" | "postsynapses" | str | list, default=True

                        Plot connectors. This can either be `True` (plot all
                        connectors), `"presynapses"` (only presynaptic connectors)
                        or `"postsynapses"` (only postsynaptic connectors). If
                        a string or a list is provided, it will be used to filter the
                        `type` column in the connectors table.

    connectors_only :   boolean, default=False

                        Plot only connectors, not the neuron.

    cn_size :           int | float, default = 1

                        Size of connectors.

    cn_layout :         dict, default={}

                        Defines default settings (color, style) for connectors.
                        See `navis.config.default_connector_colors` for the
                        default layout.

    cn_colors :         str | tuple | dict | "neuron"

                        Overrides the default connector (e.g. synpase) colors:
                            - single color as str (e.g. `'red'`) or rgb tuple
                            (e.g. `(1, 0, 0)`)
                            - dict mapping the connectors tables `type` column to
                            a color (e.g. `{"pre": (1, 0, 0)}`)
                            - with "neuron", connectors will receive the same color
                            as their neuron

    cn_mesh_colors :    bool, default=False

                        If True, will use the neuron's color for its connectors.

    scatter_kws :       dict, default={}

                        Parameters to be used when plotting points. Accepted
                        keywords are: `size` and `color`.

    volume_outlines :   bool | "both", default=False

                        If True will plot volume outline with no fill. Only
                        works with `method="2d"`. Requires the `shapely` package.

    dps_scale_vec :     float

                        Scale vector for dotprops.

    Figure parameters
    -----------------
    method :            '2d' | '3d' (default) | '3d_complex'

                        Method used to generate plot. Comes in three flavours:
                         1. `2d` uses normal matplotlib. Neurons are plotted on
                            top of one another in the order their are passed to
                            the function. Use the `view` parameter (below) to
                            set the view (default = xy).
                         2. `3d` uses matplotlib's 3D axis. Here, matplotlib
                            decide the depth order (zorder) of plotting. Can
                            change perspective either interacively or by code
                            (see examples).
                         3. `3d_complex` same as 3d but each neuron segment is
                            added individually. This allows for more complex
                            zorders to be rendered correctly. Slows down
                            rendering!

    view :              tuple, default = ("x", "y")

                        Sets view for `method='2d'`. Can be any combination of
                        "x", "y", "z" and their negations. For example, to plot
                        from the top, use `view=('x', '-y')`. For 3D `methods`,
                        this will set the initial view which can be changed by
                        adjusting `ax.azim`, `ax.elev` and `ax.roll` (see examples).

    non_view_axes3d :   "show" | "hide" (default) | "fade"

                        Only relevant for methods '3d' and '3d_complex': what to
                        do with the axis that are not in the view. If 'hide', will
                        hide them. If 'show', will show them. If 'fade', will
                        make them semi-transparent. This is relevant if you
                        intend if you intend to customize the view after plotting.

    autoscale :         bool, default=True

                        If True, will scale the axes to fit the data.

    scalebar :          int | float | str | pint.Quantity | dict, default=False

                        Adds a scale bar. Provide integer, float or str to set
                        size of scalebar. Int|float are assumed to be in same
                        units as data. You can specify units in as string:
                        e.g. "1 um". For methods '3d' and '3d_complex', this
                        will create an axis object.

                        You can customize the scalebar by passing a dictionary.
                        For example:

                        `{size: "1 micron", color: 'k', lw: 3, alpha: 0.9}`


    ax :                matplotlib.Axes, default=None

                        Pass an axis object if you want to plot on an existing
                        canvas. Must match `method` - i.e. 2D or 3D axis.

    figsize :           tuple, default=None

                        Size of figure. Ignored if `ax` is provided.

    rasterize :         bool, default=False

                        Neurons produce rather complex vector graphics which can
                        lead to large files when saving to SVG, PDF or PS. Use
                        this parameter to rasterize neurons and meshes/volumes
                        (but not axes or labels) to reduce file size.

    orthogonal :        bool, default=True

                        Whether to use orthogonal or perspective view for
                        methods '3d' and '3d_complex'.

    group_neurons :     bool, default=False

                        If True, neurons will be grouped. Works with SVG export
                        but not PDF. Does NOT work with `method='3d_complex'`.

    Returns
    -------
    fig :               matplotlib.Figure
    ax :                matplotlib.Axes

    Examples
    --------

    >>> import navis
    >>> import matplotlib.pyplot as plt

    Plot list of neurons as simple 2d:

    >>> nl = navis.example_neurons()
    >>> fig, ax = navis.plot2d(nl, method='2d', view=('x', '-z'))
    >>> plt.show() # doctest: +SKIP

    Add a volume:

    >>> vol = navis.example_volume('LH')
    >>> fig, ax = navis.plot2d([nl, vol], method='2d', view=('x', '-z'))
    >>> plt.show() # doctest: +SKIP

    Change neuron colors:

    >>> fig, ax = navis.plot2d(
    ...              nl,
    ...              method='2d',
    ...              view=('x', '-z'),
    ...              color=['r', 'g', 'b', 'm', 'c', 'y']
    ...          )
    >>> plt.show() # doctest: +SKIP

    Plot in "fake" 3D:

    >>> fig, ax = navis.plot2d(nl, method='3d', view=('x', '-z'))
    >>> plt.show() # doctest: +SKIP
    >>> # In an interactive window you can dragging the plot to rotate

    Plot in "fake" 3D and change perspective:

    >>> fig, ax = navis.plot2d(nl, method='3d', view=('x', '-z'))
    >>> # Change view
    >>> ax.elev = -20
    >>> ax.azim = 45
    >>> ax.roll = 180
    >>> plt.show() # doctest: +SKIP

    Plot using depth-coloring:

    >>> fig, ax = navis.plot2d(nl, method='3d', depth_coloring=True, view=('x', '-z'))
    >>> plt.show() # doctest: +SKIP

    See the [plotting intro](../../generated/gallery/1_plotting/tutorial_plotting_00_intro)
    for more examples.

    See Also
    --------
    [`navis.plot3d`][]
            Use this if you want interactive, perspectively correct renders
            and if you don't need vector graphics as outputs.
    [`navis.plot1d`][]
            A nifty way to visualise neurons in a single dimension.
    [`navis.plot_flat`][]
            Plot neurons as flat structures (e.g. dendrograms).

    """
    # This handles (1) checking for invalid arguments, (2) setting defaults and
    # (3) synonyms
    settings = Matplotlib2dSettings().update_settings(**kwargs)

    _METHOD_OPTIONS = ["2d", "3d", "3d_complex"]
    if settings.method not in _METHOD_OPTIONS:
        raise ValueError(
            f'Unknown method "{settings.method}". Please use either: '
            f'{",".join(_METHOD_OPTIONS)}'
        )

    # Parse objects
    (neurons, volumes, points, _) = utils.parse_objects(x)

    # Here we check whether `color_by` is a neuron property which we
    # want to translate into a single color per neuron, or a
    # per node/vertex property which we will parse late
    color_neurons_by = None
    if settings.color_by is not None and neurons:
        if not settings.palette:
            raise ValueError(
                'Must provide palette (via e.g. `palette="viridis"`) '
                "when using `color_by` argument."
            )

        # Check if this may be a neuron property
        if isinstance(settings.color_by, str):
            # Check if this could be a neuron property
            has_prop = hasattr(neurons[0], settings.color_by)

            # For TreeNeurons, we also check if it is a node property
            # If so, prioritize this.
            if isinstance(neurons[0], core.TreeNeuron):
                if settings.color_by in neurons[0].nodes.columns:
                    has_prop = False

            if has_prop:
                # If it is, use it to color neurons
                color_neurons_by = [
                    getattr(neuron, settings.color_by) for neuron in neurons
                ]
                settings.color_by = None
        elif isinstance(settings.color_by, (list, np.ndarray)):
            if len(settings.color_by) == len(neurons):
                color_neurons_by = settings.color_by
                settings.color_by = None

    # Generate the per-neuron colors
    (neuron_cmap, volumes_cmap) = prepare_colormap(
        settings.color,
        neurons=neurons,
        volumes=volumes,
        palette=settings.palette,
        color_by=color_neurons_by,
        alpha=settings.alpha,
        color_range=1,
    )

    if not isinstance(settings.color_by, type(None)):
        neuron_cmap = vertex_colors(
            neurons,
            by=settings.color_by,
            use_alpha=False,
            palette=settings.palette,
            norm_global=settings.norm_global,
            vmin=settings.vmin,
            vmax=settings.vmax,
            na="raise",
            color_range=1,
        )

    if not isinstance(settings.shade_by, type(None)):
        alphamap = vertex_colors(
            neurons,
            by=settings.shade_by,
            use_alpha=True,
            palette="viridis",  # palette is irrelevant here
            norm_global=settings.norm_global,
            vmin=settings.smin,
            vmax=settings.smax,
            na="raise",
            color_range=1,
        )

        new_colormap = []
        for c, a in zip(neuron_cmap, alphamap):
            if not (isinstance(c, np.ndarray) and c.ndim == 2):
                c = np.tile(c, (a.shape[0], 1))

            if c.shape[1] == 4:
                c[:, 3] = a[:, 3]
            else:
                c = np.insert(c, 3, a[:, 3], axis=1)

            new_colormap.append(c)
        neuron_cmap = new_colormap

    # Generate axes
    if not settings.ax:
        if settings.method == "2d":
            fig, ax = plt.subplots(figsize=settings.figsize)
        elif settings.method in ("3d", "3d_complex"):
            fig = plt.figure(
                figsize=settings.figsize if settings.figsize else plt.figaspect(1) * 1.5
            )
            ax = fig.add_subplot(111, projection="3d")
        # Hide axes
        # ax.set_axis_off()
    else:
        # Check if correct axis were provided
        if not isinstance(settings.ax, mpl.axes.Axes):
            raise TypeError('Ax must be of type "mpl.axes.Axes", ' f'not "{type(ax)}"')
        ax = settings.ax
        fig = ax.get_figure()
        if settings.method in ("3d", "3d_complex") and ax.name != "3d":
            raise TypeError("Axis must be 3d.")
        elif settings.method == "2d" and ax.name == "3d":
            raise TypeError("Axis must be 2d.")

    # Set axis projection
    if settings.method in ("3d", "3d_complex"):
        # This sets the view
        _set_view3d(ax, settings)

        # Some styling:
        # Make background transparent (nicer for dark themes)
        fig.patch.set_alpha(0)
        ax.patch.set_alpha(0)

        # For 3d axes, we also need to set the pane color to transparent
        if hasattr(ax, "zaxis"):
            ax.xaxis.pane.fill = False
            ax.xaxis.pane.set_edgecolor((1, 1, 1, 0))

            ax.yaxis.pane.fill = False
            ax.yaxis.pane.set_edgecolor((1, 1, 1, 0))

            ax.zaxis.pane.set_edgecolor((1, 1, 1, 0))
            ax.zaxis.pane.fill = False

        if settings.orthogonal:
            ax.set_proj_type("ortho")
        else:
            ax.set_proj_type("persp", focal_length=1)  # smaller = more perspective
    else:
        ax.set_aspect("equal")
        _set_view2d(ax, settings)

    # Prepare some stuff for depth coloring
    if settings.depth_coloring and not neurons.empty:
        if settings.method == "3d_complex":
            raise Exception(
                f'Depth coloring unavailable for method "{settings.method}"'
            )
        elif settings.method == "2d":
            bbox = neurons.bbox
            # Add to kwargs
            xy = [v.replace("-", "").replace("+", "") for v in settings.view]
            depth_ix = [v[1] for v in [("x", 0), ("y", 1), ("z", 2)] if v[0] not in xy]

            # We use this to track the normaliser
            settings.norm = plt.Normalize(
                vmin=bbox[depth_ix, 0], vmax=bbox[depth_ix, 1]
            )

    # Plot volumes first
    if volumes:
        for i, v in enumerate(volumes):
            _ = _plot_volume(v, volumes_cmap[i], ax, settings)

    # Create lines from segments
    visuals = {}
    for i, neuron in enumerate(
        config.tqdm(
            neurons,
            desc="Plot neurons",
            leave=False,
            disable=config.pbar_hide | len(neurons) <= 10,
        )
    ):
        if not settings.connectors_only:
            if isinstance(neuron, core.TreeNeuron) and neuron.nodes.empty:
                logger.warning(f"Skipping TreeNeuron w/o nodes: {neuron.label}")
                continue
            if isinstance(neuron, core.TreeNeuron) and neuron.nodes.shape[0] == 1:
                logger.warning(f"Skipping single-node TreeNeuron: {neuron.label}")
                continue
            elif isinstance(neuron, core.MeshNeuron) and neuron.faces.size == 0:
                logger.warning(f"Skipping MeshNeuron w/o faces: {neuron.label}")
                continue
            elif isinstance(neuron, core.Dotprops) and neuron.points.size == 0:
                logger.warning(f"Skipping Dotprops w/o points: {neuron.label}")
                continue

            if isinstance(neuron, core.TreeNeuron) and settings.radius == "auto":
                # Number of nodes with radii
                n_radii = (
                    neuron.nodes.get("radius", pd.Series([])).fillna(0) > 0
                ).sum()
                # If less than 30% of nodes have a radius, we will fall back to lines
                if n_radii / neuron.nodes.shape[0] < 0.3:
                    settings.radius = False

            if isinstance(neuron, core.TreeNeuron) and settings.radius:
                # Warn once if more than 5% of nodes have missing radii
                if not getattr(fig, "_radius_warned", False):
                    if (
                        (neuron.nodes.radius.fillna(0).values <= 0).sum()
                        / neuron.n_nodes
                    ) > 0.05:
                        logger.warning(
                            "Some skeleton nodes have radius <= 0. This may lead to "
                            "rendering artifacts. Set `radius=False` to plot skeletons "
                            "as single-width lines instead."
                        )
                        fig._radius_warned = True

                _neuron = conversion.tree2meshneuron(
                    neuron,
                    warn_missing_radii=False,
                    radius_scale_factor=settings.get("linewidth", 1),
                )
                _neuron.connectors = neuron.connectors
                neuron = _neuron

                # See if we need to map colors to vertices
                if isinstance(neuron_cmap[i], np.ndarray) and neuron_cmap[i].ndim == 2:
                    neuron_cmap[i] = neuron_cmap[i][neuron.vertex_map]

            if isinstance(neuron, core.TreeNeuron):
                lc, sc = _plot_skeleton(neuron, neuron_cmap[i], ax, settings)
                # Keep track of visuals related to this neuron
                visuals[neuron] = {"skeleton": lc, "somata": sc}
            elif isinstance(neuron, core.MeshNeuron):
                m = _plot_mesh(neuron, neuron_cmap[i], ax, settings)
                visuals[neuron] = {"mesh": m}
            elif isinstance(neuron, core.Dotprops):
                dp = _plot_dotprops(neuron, neuron_cmap[i], ax, settings)
                visuals[neuron] = {"dotprop": dp}
            elif isinstance(neuron, core.VoxelNeuron):
                dp = _plot_voxels(
                    neuron,
                    neuron_cmap[i],
                    ax,
                    settings,
                    **settings.scatter_kws,
                )
                visuals[neuron] = {"dotprop": dp}
            else:
                raise TypeError(
                    f"Don't know how to plot neuron of type '{type(neuron)}' "
                )

        if (settings.connectors or settings.connectors_only) and neuron.has_connectors:
            _ = _plot_connectors(neuron, neuron_cmap[i], ax, settings)

    # Plot points
    for p in points:
        _ = _plot_scatter(p, ax, settings)

    # Note: autoscaling is a bitch for 3d. In particular when we use Collections, because
    # these are currently ignored by matplotlib's built-in autoscaling.
    if settings.autoscale:
        ax.autoscale(tight=False)  # tight=False avoids clipping the neurons

        if "3d" in settings.method:
            update_axes3d_bounds(ax)

        # This is apparently still required and has to happen AFTER updating axis bounds
        ax.set_aspect("equal", adjustable="box")

    # Add scalebar after the dust has settled
    if settings.scalebar not in (False, None):
        if not settings.orthogonal:
            raise ValueError("Scalebar only available if `orthogonal=True`.")

        _ = _add_scalebar(settings.scalebar, neurons, ax, settings)

    def set_depth():
        """Set depth information for neurons according to camera position."""
        # Get projected coordinates
        proj_co = proj_points(all_co, ax.get_proj())

        # Get min and max of z coordinates
        z_min, z_max = min(proj_co[:, 2]), max(proj_co[:, 2])

        # Generate a new normaliser
        norm = plt.Normalize(vmin=z_min, vmax=z_max)

        # Go over all neurons and update Z information
        for neuron in visuals:
            # Get this neurons colletion and coordinates
            if "skeleton" in visuals[neuron]:
                c = visuals[neuron]["skeleton"]
                this_co = c._segments3d[:, 0, :]
            elif "mesh" in visuals[neuron]:
                c = visuals[neuron]["mesh"]
                # Note that we only get every third position -> that's because
                # these vectors actually represent faces, i.e. each vertex
                this_co = c._vec.T[::3, [0, 1, 2]]
            else:
                raise ValueError(
                    f"Neither mesh nor skeleton found for neuron {neuron.id}"
                )

            # Get projected coordinates
            this_proj = proj_points(this_co, ax.get_proj())

            # Normalise z coordinates
            ns = norm(this_proj[:, 2]).data

            # Set array
            c.set_array(ns)

            # No need for normaliser - already happened
            c.set_norm(None)

            if isinstance(neuron, core.TreeNeuron) and not isinstance(
                getattr(neuron, "soma", None), type(None)
            ):
                # Get depth of soma(s)
                soma = utils.make_iterable(neuron.soma)
                soma_co = (
                    neuron.nodes.set_index("node_id").loc[soma][["x", "y", "z"]].values
                )
                soma_proj = proj_points(soma_co, ax.get_proj())
                soma_cs = norm(soma_proj[:, 2]).data

                # Set soma color
                for cs, s in zip(soma_cs, visuals[neuron]["somata"]):
                    s.set_color(cmap(cs))

    def Update(event):
        set_depth()

    if settings.depth_coloring:
        if settings.palette:
            cmap = plt.get_cmap(settings.palette)
        else:
            cmap = DEPTH_CMAP
        if settings.method == "2d" and settings.depth_scale:
            sm = ScalarMappable(norm=settings.norm, cmap=cmap)
            fig.colorbar(sm, ax=ax, fraction=0.075, shrink=0.5, label="Depth")
        elif settings.method == "3d":
            # Collect all coordinates
            all_co = []
            for n in visuals:
                if "skeleton" in visuals[n]:
                    all_co.append(visuals[n]["skeleton"]._segments3d[:, 0, :])
                if "mesh" in visuals[n]:
                    all_co.append(visuals[n]["mesh"]._vec.T[:, [0, 1, 2]])

            all_co = np.concatenate(all_co, axis=0)
            fig.canvas.mpl_connect("draw_event", Update)
            set_depth()

    return fig, ax


def _add_scalebar(scalebar, neurons, ax, settings):
    """Add scalebar."""
    defaults = {
        "color": "black",
        "lw": 3,
        "alpha": 0.9,
    }

    if isinstance(scalebar, dict):
        if "size" not in scalebar:
            raise ValueError("`scalebar` dictionary must contain 'size' key.")
        defaults.update(scalebar)
        scalebar = defaults["size"]

    if isinstance(scalebar, bool):
        scalebar = "1 um"

    if isinstance(scalebar, str):
        scalebar = config.ureg(scalebar)

    if isinstance(scalebar, pint.Quantity):
        # If we have neurons as points of reference convert
        if neurons:
            scalebar = scalebar.to(neurons[0].units).magnitude
        # If no reference, use assume it's the same units
        else:
            scalebar = scalebar.magnitude

    # Hard-coded 5% offset from figure boundaries
    ax_offset = (ax.get_xlim()[1] - ax.get_xlim()[0]) / 100 * 5

    if settings.method == "2d":
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        coords = np.array([[xlim[0], ylim[0]], [xlim[0] + scalebar, ylim[0]]])

        if not ax.xaxis.get_inverted():
            coords[:, 0] += ax_offset
        else:
            coords[:, 0] -= ax_offset

        if not ax.yaxis.get_inverted():
            coords[:, 1] += ax_offset
        else:
            coords[:, 1] -= ax_offset

        sbar = mlines.Line2D(
            coords[:, 0],
            coords[:, 1],
            lw=defaults["lw"],
            alpha=defaults["alpha"],
            color=defaults["color"],
            zorder=1000,
        )
        sbar.set_gid(f"{scalebar}_scalebar")

        ax.add_line(sbar)
    elif settings.method in ["3d", "3d_complex"]:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        zlim = ax.get_zlim()

        left = xlim[0] + ax_offset
        bottom = zlim[0] + ax_offset
        front = ylim[0] + ax_offset

        sbar = [
            np.array([[left, front, bottom], [left, front, bottom]]),
            np.array([[left, front, bottom], [left, front, bottom]]),
            np.array([[left, front, bottom], [left, front, bottom]]),
        ]
        sbar[0][1][0] += scalebar
        sbar[1][1][1] += scalebar
        sbar[2][1][2] += scalebar

        lc = Line3DCollection(
            sbar, color=defaults["color"], lw=defaults["lw"], alpha=defaults["alpha"]
        )
        lc.set_gid(f"{scalebar}_scalebar")

        ax.add_collection3d(lc, autolim=False)


def _plot_scatter(points, ax, settings):
    """Plot dotprops."""
    if settings.method == "2d":
        default_settings = dict(c="black", zorder=4, edgecolor="none", s=1)
        default_settings.update(settings.scatter_kws)
        default_settings = _fix_default_dict(default_settings)

        x, y = _parse_view2d(points, settings.view)

        ax.scatter(x, y, **default_settings)
    elif settings.method in ["3d", "3d_complex"]:
        default_settings = dict(c="black", s=1, depthshade=False, edgecolor="none")
        default_settings.update(settings.scatter_kws)
        default_settings = _fix_default_dict(default_settings)

        ax.scatter(points[:, 0], points[:, 1], points[:, 2], **default_settings)


def _plot_voxels(vx, color, ax, settings, **scatter_kws):
    """Plot VoxelNeuron as scatter plot."""
    # Use only the top N voxels
    assert isinstance(vx, core.VoxelNeuron)
    n_pts = 1000000
    v = vx.values
    pts = vx.voxels
    srt = np.argsort(v)[::-1]

    pts = pts[srt][:n_pts]
    v = v[srt][:n_pts]

    # Scale points by units
    pts = pts * vx.units_xyz.magnitude + vx.offset

    # Calculate colors
    cmap = color_to_cmap(color)
    colors = cmap(v / v.max())

    if settings.method == "2d":
        x, y = _parse_view2d(pts, settings.view)
        ax.scatter(x, y, c=colors, s=scatter_kws.get("size", 20))
    elif settings.method in ["3d", "3d_complex"]:
        ax.scatter(
            pts[:, 0],
            pts[:, 1],
            pts[:, 2],
            c=colors,
            marker=scatter_kws.get("marker", "o"),
            s=scatter_kws.get("size", 0.1),
        )


def color_to_cmap(color):
    """Convert single color to color palette."""
    color = mcl.to_rgb(color)

    colors = [[color[0], color[1], color[2], 0], [color[0], color[1], color[2], 1]]

    return mcl.LinearSegmentedColormap.from_list("Palette", colors, N=256)


def _plot_dotprops(dp, color, ax, settings):
    """Plot dotprops."""
    # Here, we will effectively cheat and turn the dotprops into a skeleton
    # which we can then pass to _plot_skeleton
    tn = dp.to_skeleton(scale_vec=settings.dps_scale_vec)

    return _plot_skeleton(tn, color, ax, settings)


def _plot_connectors(neuron, color, ax, settings):
    """Plot connectors."""
    cn_layout = copy.deepcopy(config.default_connector_colors)

    if settings.connectors == "pre":
        connectors = neuron.presynapses
    elif settings.connectors == "post":
        connectors = neuron.postsynapses
    elif isinstance(settings.connectors, str):
        connectors = neuron.connectors[neuron.connectors.type == settings.connectors]
    elif isinstance(settings.connectors, (list, np.ndarray, tuple)):
        connectors = neuron.connectors[neuron.connectors.type.isin(settings.connectors)]
    else:
        connectors = neuron.connectors

    if connectors.empty:
        return

    # Update with user settings
    if settings.cn_layout:
        cn_layout.update(settings.cn_layout)

    # Update with user color
    if settings.cn_mesh_colors or settings.cn_layout == "neuron":
        # change all of the colors to color
        for inner_dict in cn_layout.values():
            # Skip non-color settings
            if not isinstance(inner_dict, dict):
                continue
            inner_dict["color"] = color

    if settings.cn_colors:
        if isinstance(settings.cn_colors, dict):
            cn_layout.update(settings.cn_colors)
        else:
            for inner_dict in cn_layout.values():
                # Skip non-color settings
                if not isinstance(inner_dict, dict):
                    continue
                inner_dict["color"] = settings.cn_colors

    if settings.method == "2d":
        for c, this_cn in connectors.groupby("type"):
            x, y = _parse_view2d(this_cn[["x", "y", "z"]].values, settings.view)

            ax.scatter(
                x,
                y,
                color=cn_layout[c]["color"],
                edgecolor="none",
                s=settings.cn_size if settings.cn_size else cn_layout["size"],
                zorder=1000,
            )
            ax.get_children()[-1].set_gid(f"CN_{neuron.id}")
    elif settings.method in ["3d", "3d_complex"]:
        c = [cn_layout[i]["color"] for i in connectors.type.values]
        ax.scatter(
            connectors.x.values,
            connectors.y.values,
            connectors.z.values,
            color=c,
            s=settings.cn_size if settings.cn_size else cn_layout["size"],
            depthshade=cn_layout.get("depthshade", False),
            zorder=0,
            edgecolor="none",
        )
        ax.get_children()[-1].set_gid(f"CN_{neuron.id}")


def _plot_mesh(neuron, color, ax, settings):
    """Plot mesh (i.e. MeshNeuron)."""
    # Map vertex colors to faces (if need be)
    if isinstance(color, np.ndarray) and color.ndim == 2:
        if len(color) != len(neuron.faces) and len(color) == len(neuron.vertices):
            color = np.array([color[f].mean(axis=0)[:4].tolist() for f in neuron.faces])

    ts = None
    if settings.method == "2d":
        # Generate 2d representation
        xy = np.dstack(_parse_view2d(neuron.vertices, settings.view))[0]

        pc = PolyCollection(
            xy[neuron.faces],
            linewidth=0,
            rasterized=settings.rasterize,
            edgecolor="none",
            label=getattr(neuron, "name"),
            zorder=100,  # unless we set this, the mesh will always be behind the grid
        )

        if settings.depth_coloring:
            if settings.palette:
                cmap = plt.get_cmap(settings.palette)
            else:
                cmap = DEPTH_CMAP

            pc.set_cmap(cmap)
            pc.set_norm(settings.norm)
            pc.set_alpha(settings.alpha if isinstance(settings.alpha, float) else None)

            # Get face centers
            if hasattr(neuron, "trimesh"):
                centers = neuron.trimesh.triangles_center[
                    :, _get_depth_axis(settings.view)
                ]
            else:
                centers = neuron.triangles_center[:, _get_depth_axis(settings.view)]
            # Set face centers to color the scale
            pc.set_array(centers)
        else:
            pc.set_facecolor(color)

        ax.add_collection(pc)
    else:
        ts = ax.plot_trisurf(
            neuron.vertices[:, 0],
            neuron.vertices[:, 1],
            neuron.faces,
            neuron.vertices[:, 2],
            label=getattr(neuron, "name"),
            rasterized=settings.rasterize,
            shade=settings.mesh_shade,
        )

        if settings.depth_coloring:
            if settings.palette:
                cmap = plt.get_cmap(settings.palette)
            else:
                cmap = DEPTH_CMAP
            ts.set_cmap(cmap)
            ts.set_alpha(settings.alpha)
        else:
            ts.set_facecolor(color)

        if settings.group_neurons:
            ts.set_gid(neuron.id)
    return ts


def _get_depth_axis(view):
    """Return index of axis which is not used for x/y."""
    view = [v.replace("-", "").replace("+", "") for v in view]
    depth = [ax for ax in ["x", "y", "z"] if ax not in view][0]
    map = {"x": 0, "y": 1, "z": 2}
    return map[depth]


def _parse_view2d(co, view):
    """Parse view parameter and returns x/y parameter."""
    if not isinstance(co, np.ndarray):
        co = np.array(co)

    map = {"x": 0, "y": 1, "z": 2}

    x_ix = map[view[0].replace("-", "").replace("+", "")]
    y_ix = map[view[1].replace("-", "").replace("+", "")]

    if co.ndim == 2:
        x = co[:, x_ix]
        y = co[:, y_ix]

        # Do NOT remove the list() here - for some reason the multiplication
        # above causes issues in matplotlib
        return (list(x), list(y))
    elif co.ndim == 3:
        xy = co[:, :, [x_ix, y_ix]]
        return xy
    else:
        raise ValueError(f"Expect coordinates to have 2 or 3 dimensions, got {co.ndim}")


def _set_view2d(ax, settings):
    """Set the axes based on the view parameter."""
    if settings.view[0].startswith("-") and not ax.xaxis.get_inverted():
        ax.invert_xaxis()
    if settings.view[1].startswith("-") and not ax.yaxis.get_inverted():
        ax.invert_yaxis()

    ax.set_xlabel(settings.view[0].replace("-", ""))
    ax.set_ylabel(settings.view[1].replace("-", ""))

    ax.grid()


def _set_view3d(ax, settings):
    """Parse view parameter into azimuth, elevation and roll for the camera."""
    # `view` can be e.g. ("x", "y"), ("x", "-y") or ("x", "z")
    # We need to convert this into azimuth, elevation and roll
    # Azimuth is the angle of the view in the x-y plane
    # Elevation is the angle of the view from the x-y plane
    # Roll is the rotation of the view around the z-axis
    view = tuple(settings.view)

    views = {
        ("x", "y"): (90, -90, 0),
        ("-x", "-y"): (90, 90, 0),
        ("x", "-y"): (-90, 90, 180),
        ("-x", "y"): (-90, 90, 0),
        ("x", "z"): (0, -90, 0),
        ("-x", "z"): (0, 90, 0),
        ("x", "-z"): (0, 90, -180),
        ("-x", "-z"): (0, 90, 90),
        ("y", "z"): (0, 0, 0),
        ("y", "-z"): (180, 0, 0),
        ("-y", "-z"): (0, 0, 180),
        ("-y", "z"): (180, 0, 180),
        ("z", "y"): (0, 0, -90),
        ("-z", "-y"): (0, 0, 90),
        ("-z", "y"): (180, 180, -90),
        ("z", "-y"): (180, 180, 90),
        # TODO: add (z, x) and (y, x) views
    }
    if view not in views:
        raise ValueError(
            f"View {view} not possible without flipping data. Please choose from {views.keys()}"
        )

    # Set view
    ax.view_init(*views[view])

    # This both sets the aspect ratio as well as zooming in slightly
    # Note: we do not have to use ax.set_aspect("equal") again
    ax.set_box_aspect([1, 1, 1], zoom=1.2)

    # Set aspect ratio
    # ax.set_aspect('equal', adjustable='box')

    # Set labels in case somebody unhides the axis
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    non_view_axis = [
        v for v in ["x", "y", "z"] if v not in [v.replace("-", "") for v in view]
    ][0]
    if settings.non_view_axes3d == "hide":
        getattr(ax, f"set_{non_view_axis}label")("")
        getattr(ax, f"set_{non_view_axis}ticks")([])
    elif settings.non_view_axes3d == "fade":
        getattr(ax, f"set_{non_view_axis}label")(non_view_axis, alpha=0.5)
        getattr(ax, f"set_{non_view_axis}ticks")([])


def _plot_skeleton(neuron, color, ax, settings):
    """Plot skeleton."""

    if settings.method == "2d":
        if not settings.depth_coloring and not (
            isinstance(color, np.ndarray) and color.ndim == 2
        ):
            # Generate by-segment coordinates
            coords = segments_to_coords(neuron, modifier=(1, 1, 1))

            # We have to add (None, None, None) to the end of each
            # slab to make that line discontinuous there
            coords = np.vstack([np.append(t, [[None] * 3], axis=0) for t in coords])

            x, y = _parse_view2d(coords, settings.view)
            this_line = mlines.Line2D(
                x,
                y,
                lw=settings.linewidth,
                ls=settings.linestyle,
                color=color,
                rasterized=settings.rasterize,
                label=f'{getattr(neuron, "name", "NA")} - #{neuron.id}',
            )
            ax.add_line(this_line)
        else:
            if isinstance(settings.palette, str):
                cmap = plt.get_cmap(settings.palette)
            else:
                cmap = DEPTH_CMAP

            coords = tn_pairs_to_coords(neuron, modifier=(1, 1, 1))
            xy = _parse_view2d(coords, settings.view)
            lc = LineCollection(
                xy,
                cmap=cmap if settings.depth_coloring else None,
                norm=settings.norm if settings.depth_coloring else None,
                rasterized=settings.rasterize,
                joinstyle="round",
            )

            lc.set_linewidth(settings.linewidth)
            lc.set_linestyle(settings.linestyle)
            lc.set_label(f'{getattr(neuron, "name", "NA")} - #{neuron.id}')

            if settings.depth_coloring:
                lc.set_array(
                    neuron.nodes.loc[neuron.nodes.parent_id >= 0][
                        ["x", "y", "z"]
                    ].values[:, _get_depth_axis(settings.view)]
                )
            elif isinstance(color, np.ndarray) and color.ndim == 2:
                # If we have a color for each node, we need to drop the roots
                if color.shape[1] != coords.shape[0]:
                    lc.set_color(color[neuron.nodes.parent_id.values >= 0])
                else:
                    lc.set_color(color)

            ax.add_collection(lc)

        if settings.soma and np.any(neuron.soma):
            soma = utils.make_iterable(neuron.soma)
            # If soma detection is messed up we might end up producing
            # dozens of soma which will freeze the kernel
            if len(soma) >= 10:
                logger.warning(f"{neuron.id} - {len(soma)} somas found.")
            for s in soma:
                if isinstance(color, np.ndarray) and color.ndim > 1:
                    s_ix = np.where(neuron.nodes.node_id == s)[0][0]
                    soma_color = color[s_ix]
                else:
                    soma_color = color

                n = neuron.nodes.set_index("node_id").loc[s]
                r = (
                    getattr(n, neuron.soma_radius)
                    if isinstance(neuron.soma_radius, str)
                    else neuron.soma_radius
                )

                if settings.depth_coloring:
                    d = [n.x, n.y, n.z][_get_depth_axis(settings.view)]
                    soma_color = DEPTH_CMAP(settings.norm(d))

                soma_defaults = dict(
                    radius=r,
                    fill=True,
                    fc=soma_color,
                    rasterized=settings.rasterize,
                    zorder=4,
                    edgecolor="none",
                )
                if isinstance(settings.soma, dict):
                    soma_defaults.update(settings.soma)

                sx, sy = _parse_view2d(np.array([[n.x, n.y, n.z]]), settings.view)
                c = mpatches.Circle((sx[0], sy[0]), **soma_defaults)
                ax.add_patch(c)
        return None, None

    elif settings.method in ["3d", "3d_complex"]:
        # For simple scenes, add whole neurons at a time to speed up rendering
        if settings.method == "3d":
            if (
                isinstance(color, np.ndarray) and color.ndim == 2
            ) or settings.depth_coloring:
                coords = tn_pairs_to_coords(neuron, modifier=(1, 1, 1))
                # If we have a color for each node, we need to drop the roots
                if isinstance(color, np.ndarray) and color.shape[1] != coords.shape[0]:
                    line_color = color[neuron.nodes.parent_id.values >= 0]
                else:
                    line_color = color
            else:
                # Generate by-segment coordinates
                coords = segments_to_coords(neuron, modifier=(1, 1, 1))
                line_color = color

            if settings.palette:
                cmap = plt.get_cmap(settings.palette)
            else:
                cmap = DEPTH_CMAP

            lc = Line3DCollection(
                coords,
                color=line_color if not settings.depth_coloring else None,
                label=neuron.id,
                cmap=cmap if settings.depth_coloring else None,
                lw=settings.linewidth,
                joinstyle="round",
                rasterized=settings.rasterize,
                linestyle=settings.linestyle,
            )
            if settings.group_neurons:
                lc.set_gid(neuron.id)
            # Need to get this before adding data
            line3D_collection = lc
            ax.add_collection3d(lc, autolim=False)

        # For complex scenes, add each segment as a single collection
        # -> helps reducing Z-order errors
        elif settings.method == "3d_complex":
            # Generate by-segment coordinates
            coords = segments_to_coords(neuron, modifier=(1, 1, 1))
            for c in coords:
                lc = Line3DCollection(
                    [c],
                    color=color,
                    lw=settings.linewidth,
                    rasterized=settings.rasterize,
                    linestyle=settings.linestyle,
                )
                if settings.group_neurons:
                    lc.set_gid(neuron.id)
                ax.add_collection3d(lc, autolim=False)
            line3D_collection = None

        surf3D_collections = []
        if settings.soma and not isinstance(getattr(neuron, "soma", None), type(None)):
            soma = utils.make_iterable(neuron.soma)
            # If soma detection is messed up we might end up producing
            # dozens of soma which will freeze the kernel
            if len(soma) >= 5:
                logger.warning(
                    f"Neuron {neuron.id} appears to have {len(soma)}"
                    " somas. Skipping plotting its somas."
                )
            else:
                for s in soma:
                    if isinstance(color, np.ndarray) and color.ndim > 1:
                        s_ix = np.where(neuron.nodes.node_id == s)[0][0]
                        soma_color = color[s_ix]
                    else:
                        soma_color = color

                    n = neuron.nodes.set_index("node_id").loc[s]
                    r = (
                        getattr(n, neuron.soma_radius)
                        if isinstance(neuron.soma_radius, str)
                        else neuron.soma_radius
                    )

                    resolution = 20
                    u = np.linspace(0, 2 * np.pi, resolution)
                    v = np.linspace(0, np.pi, resolution)
                    x = r * np.outer(np.cos(u), np.sin(v)) + n.x
                    y = r * np.outer(np.sin(u), np.sin(v)) + n.y
                    z = r * np.outer(np.ones(np.size(u)), np.cos(v)) + n.z

                    soma_defaults = dict(
                        color=soma_color,
                        shade=settings.mesh_shade,
                        rasterized=settings.rasterize,
                    )
                    if isinstance(settings.soma, dict):
                        soma_defaults.update(settings.soma)

                    surf = ax.plot_surface(x, y, z, **soma_defaults)

                    if settings.group_neurons:
                        surf.set_gid(neuron.id)

                    surf3D_collections.append(surf)

        return line3D_collection, surf3D_collections


def _plot_volume(volume, color, ax, settings):
    """Plot volume."""
    name = getattr(volume, "name")

    if len(color) == 4:
        this_alpha = color[3]
    else:
        this_alpha = 1

    if settings.volume_outlines:
        fill, lw, fc, ec = False, 1, "none", color
    else:
        fill, lw, fc, ec = True, 0, color, "none"

    if settings.method == "2d":
        if settings.volume_outlines in (False, "both"):
            # Generate 2d representation
            xy = np.dstack(_parse_view2d(volume.verts, settings.view))[0]

            # Generate a patch for each face
            pc = PolyCollection(
                xy[volume.faces],
                linewidth=lw,
                facecolor=fc,
                rasterized=settings.rasterize,
                edgecolor=ec,
                alpha=this_alpha,
                zorder=0,
            )
            ax.add_collection(pc)

        if settings.volume_outlines in (True, "both"):
            verts = volume.to_2d(view=settings.view, alpha=0.001)
            vpatch = mpatches.Polygon(
                verts,
                closed=True,
                lw=lw,
                fill=fill,
                rasterized=settings.rasterize,
                fc=fc,
                ec=ec,
                zorder=0,
                alpha=1 if settings.volume_outlines == "both" else this_alpha,
            )
            ax.add_patch(vpatch)

    elif settings.method in ["3d", "3d_complex"]:
        if settings.volume_outlines:
            logger.warning("Volume outlines are not supported for 3d plotting. ")

        verts = np.vstack(volume.vertices)

        # Add alpha
        if len(color) == 3:
            color = (color[0], color[1], color[2], 0.1)

        ts = ax.plot_trisurf(
            verts[:, 0],
            verts[:, 1],
            volume.faces,
            verts[:, 2],
            label=name,
            rasterized=settings.rasterize,
            color=color,
        )
        ts.set_gid(name)


def _fix_default_dict(x):
    """Consolidate duplicate settings.

    E.g. scatter kwargs when 'c' and 'color' is provided.

    """
    # The first entry is the "survivor"
    duplicates = [["color", "c"], ["size", "s"], ["alpha", "a"]]

    for dupl in duplicates:
        if sum([v in x for v in dupl]) > 1:
            to_delete = [v for v in dupl if v in x][1:]
            _ = [x.pop(v) for v in to_delete]

    return x


def proj_points(points, M):
    """Project points using a projection matrix.

    This was previously done using the analagous function
    mpl_toolkits.mplot3d.proj3d.proj_points but that is deprecated.
    """
    xs, ys, zs = zip(*points)
    vec = np.array([xs, ys, zs, np.ones_like(xs)])

    vecw = np.dot(M, vec)
    w = vecw[3]
    # clip here..
    txs, tys, tzs = vecw[0] / w, vecw[1] / w, vecw[2] / w

    return np.column_stack((txs, tys, tzs))


def update_axes3d_bounds(ax):
    """Update axis bounds and remove default points (0,0,0) and (1,1,1)."""
    # Collect data points present in the figure
    points = []
    for c in ax.collections:
        if isinstance(c, Line3DCollection):
            for s in c._segments3d:
                points.append(s)
        elif isinstance(c, Poly3DCollection):
            points.append(c._vec[:3, :].T)
        elif isinstance(c, (Path3DCollection, Patch3DCollection)):
            points.append(np.array(c._offsets3d).T)

    if not len(points):
        return

    points = np.vstack(points)

    # If this is the first set of points, we need to overwrite the defaults
    # That should happen automatically but for some reason doesn't for 3d axes
    if not getattr(ax, "had_data", False):
        mn = points.min(axis=0)
        mx = points.max(axis=0)
        new_xybounds = np.array([[mn[0], mn[1]], [mx[0], mx[1]]])
        new_zzbounds = np.array([[mn[2], mn[2]], [mx[2], mx[2]]])
        ax.xy_dataLim.set_points(new_xybounds)
        ax.zz_dataLim.set_points(new_zzbounds)
        ax.xy_viewLim.set_points(new_xybounds)
        ax.zz_viewLim.set_points(new_zzbounds)
        ax.had_data = True
    else:
        ax.auto_scale_xyz(
            points[:, 0].tolist(),
            points[:, 1].tolist(),
            points[:, 2].tolist(),
            had_data=True,
        )



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import io

import networkx as nx
import pandas as pd
import xml.etree.ElementTree as ET

from typing import Union, Dict, Optional, Any, IO, Iterable
from zipfile import ZipFile

from .. import config, core
from . import base


__all__ = ["read_nmx", "read_nml"]

# Set up logging
logger = config.get_logger(__name__)

NODE_COLUMNS = ("node_id", "label", "x", "y", "z", "radius", "parent_id")
DEFAULT_PRECISION = 32
DEFAULT_FMT = "{name}.nmx"


class NMLReader(base.BaseReader):
    def __init__(
        self,
        precision: int = DEFAULT_PRECISION,
        attrs: Optional[Dict[str, Any]] = None,
        errors: str = "raise",
    ):
        super().__init__(
            fmt="",
            attrs=attrs,
            file_ext=".nml",
            read_binary=False,
            errors=errors,
            name_fallback="NML",
        )

        int_, float_ = base.parse_precision(precision)
        self._dtypes = {
            "node_id": int_,
            "parent_id": int_,
            "label": "category",
            "x": float_,
            "y": float_,
            "z": float_,
            "radius": float_,
        }

    @base.handle_errors
    def read_buffer(
        self, f: IO, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.TreeNeuron":
        """Read .nml buffer into a TreeNeuron.

        NML files are XML-encoded files containing data for a single neuron.

        Parameters
        ----------
        f :         IO
                    Readable buffer (must be bytes).
        attrs :     dict | None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.TreeNeuron
        """
        return self.read_nml(f.read(), attrs=attrs)

    def read_nml(
        self, f: IO, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.TreeNeuron":
        """Read .nml buffer into a TreeNeuron.

        NML files are XML files containing a single neuron.

        Parameters
        ----------
        f :         IO
                    Readable buffer.
        attrs :     dict | None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.TreeNeuron
        """
        if isinstance(f, bytes):
            f = f.decode()

        f = io.StringIO(f)
        root = ET.parse(f).getroot()

        # Copy the attributes dict
        for element in root:
            if element.tag == "thing":
                nodes = pd.DataFrame.from_records([n.attrib for n in element[0]])
                edges = pd.DataFrame.from_records([n.attrib for n in element[1]])
                edges = edges.astype(self._dtypes["node_id"])

                nodes.rename({"id": "node_id"}, axis=1, inplace=True)
                nodes = nodes.astype(
                    {k: v for k, v in self._dtypes.items() if k in nodes.columns}
                )

        G = nx.Graph()
        G.add_edges_from(edges.values)
        tree = nx.bfs_tree(G, list(G.nodes)[0])
        edges = pd.DataFrame(list(tree.edges), columns=["source", "target"])
        nodes["parent_id"] = (
            edges.set_index("target").reindex(nodes.node_id.values).source.values
        )
        nodes["parent_id"] = nodes.parent_id.fillna(-1).astype(self._dtypes["node_id"])
        nodes.sort_values("node_id", inplace=True)

        return core.TreeNeuron(
            nodes, **(self._make_attributes({"name": "NML", "origin": "nml"}, attrs))
        )


class NMXReader(NMLReader):
    """This is a version of the NML file reader that reads from zipped archives."""

    def __init__(
        self,
        precision: int = DEFAULT_PRECISION,
        attrs: Optional[Dict[str, Any]] = None,
        errors: str = "raise",
    ):
        super().__init__(precision=precision, errors=errors, attrs=attrs)

        # Overwrite some of the settings
        self.read_binary = True
        self.file_ext = ".nmx"
        self.name_fallback = "NMX"

    @base.handle_errors
    def read_buffer(
        self, f: IO, attrs: Optional[Dict[str, Any]] = None
    ) -> "core.TreeNeuron":
        """Read .nmx buffer into a TreeNeuron.

        NMX files are zip files containing XML-encoded .nml files containing
        data for a single neuron.

        Parameters
        ----------
        f :         IO
                    Readable buffer (must be bytes).
        attrs :     dict | None
                    Arbitrary attributes to include in the TreeNeuron.

        Returns
        -------
        core.TreeNeuron
        """
        if not isinstance(f.read(0), bytes):
            raise ValueError(f'Expected bytes, got "{type(f.read(0))}"')

        zip = ZipFile(f)
        for f in zip.filelist:
            if f.filename.endswith(".nml") and "skeleton" in f.filename:
                attrs["file"] = f.filename
                attrs["id"] = f.filename.split("/")[0]
                return self.read_nml(zip.read(f), attrs=attrs)
        logger.warning(
            f'Skipped "{f.filename.split("/")[0]}.nmx": failed to ' 'import skeleton.'
        )


def read_nmx(
    f: Union[str, pd.DataFrame, Iterable],
    include_subdirs: bool = False,
    parallel: Union[bool, int] = "auto",
    precision: int = 32,
    limit: Optional[int] = None,
    errors: str = "raise",
    **kwargs,
) -> "core.NeuronObject":
    """Read NMX files into Neuron/Lists.

    NMX is an xml-based format used by pyKNOSSOS.
    See e.g. [here](https://doi.org/10.5281/zenodo.58985) for a data dump
    of neurons from Wanner et al. (2016).

    Parameters
    ----------
    f :                 str
                        Filename or folder. If folder, will import all `.nmx`
                        files.
    include_subdirs :   bool, optional
                        If True and `f` is a folder, will also search
                        subdirectories for `.nmx` files.
    parallel :          "auto" | bool | int
                        Defaults to `auto` which means only use parallel
                        processing if more than 200 files are imported. Spawning
                        and joining processes causes overhead and is
                        considerably slower for imports of small numbers of
                        neurons. Integer will be interpreted as the
                        number of cores (otherwise defaults to
                        `os.cpu_count() // 2`).
    precision :         int [8, 16, 32, 64] | None
                        Precision for data. Defaults to 32 bit integers/floats.
                        If `None` will let pandas infer data types - this
                        typically leads to higher than necessary precision.
    limit :             int | str | slice | list, optional
                        When reading from a folder or archive you can use this parameter to
                        restrict the which files read:
                         - if an integer, will read only the first `limit` NMX files
                           (useful to get a sample from a large library of skeletons)
                         - if a string, will interpret it as filename (regex) pattern
                           and only read files that match the pattern; e.g. `limit='.*_R.*'`
                           will only read files that contain `_R` in their filename
                         - if a slice (e.g. `slice(10, 20)`) will read only the files in
                           that range
                         - a list is expected to be a list of filenames to read from
                           the folder/archive
    errors :            "raise" | "log" | "ignore"
                        If "log" or "ignore", errors will not be raised and the
                        mesh will be skipped. Can result in empty output.
    **kwargs
                        Keyword arguments passed to the construction of
                        `navis.TreeNeuron`. You can use this to e.g. set
                        meta data.

    Returns
    -------
    navis.NeuronList

    See Also
    --------
    [`navis.read_nml`][]
                        Read NML file(s).

    """
    reader = NMXReader(precision=precision, errors=errors, attrs=kwargs)
    # Read neurons
    neurons = reader.read_any(
        f, parallel=parallel, limit=limit, include_subdirs=include_subdirs
    )

    # Failed reads will produce empty neurons which we need to remove
    if isinstance(neurons, core.NeuronList):
        neurons = neurons[neurons.has_nodes]

    return neurons


def read_nml(
    f: Union[str, pd.DataFrame, Iterable],
    include_subdirs: bool = False,
    parallel: Union[bool, int] = "auto",
    precision: int = 32,
    limit: Optional[int] = None,
    **kwargs,
) -> "core.NeuronObject":
    """Read xml-based NML files into Neuron/Lists.

    Parameters
    ----------
    f :                 str
                        Filename or folder. If folder, will import all `.nml`
                        files.
    include_subdirs :   bool, optional
                        If True and `f` is a folder, will also search
                        subdirectories for `.nml` files.
    parallel :          "auto" | bool | int
                        Defaults to `auto` which means only use parallel
                        processing if more than 200 files are imported. Spawning
                        and joining processes causes overhead and is
                        considerably slower for imports of small numbers of
                        neurons. Integer will be interpreted as the
                        number of cores (otherwise defaults to
                        `os.cpu_count() // 2`).
    precision :         int [8, 16, 32, 64] | None
                        Precision for data. Defaults to 32 bit integers/floats.
                        If `None` will let pandas infer data types - this
                        typically leads to higher than necessary precision.
    limit :             int | str | slice | list, optional
                        When reading from a folder or archive you can use this
                        parameter to restrict the which files read:
                         - if an integer, will read only the first `limit` NML files
                           (useful to get a sample from a large library of skeletons)
                         - if a string, will interpret it as filename (regex) pattern
                           and only read files that match the pattern; e.g. `limit='.*_R.*'`
                           will only read files that contain `_R` in their filename
                         - if a slice (e.g. `slice(10, 20)`) will read only the files in
                           that range
                         - a list is expected to be a list of filenames to read from
                           the folder/archive
    **kwargs
                        Keyword arguments passed to the construction of
                        `navis.TreeNeuron`. You can use this to e.g. set
                        meta data.

    Returns
    -------
    navis.NeuronList

    See Also
    --------
    [`navis.read_nmx`][]
                        Read NMX files (collections of NML files).

    """
    reader = NMLReader(precision=precision, attrs=kwargs)
    # Read neurons
    neurons = reader.read_any(
        f, parallel=parallel, limit=limit, include_subdirs=include_subdirs
    )

    return neurons



import navis
import flybrains

# Add fake bounds for JRCFIB
flybrains.JRCFIB2018Fraw.boundingbox = [0, 34499, 7409, 37539, 2952, 40076]


def test_skeleton_transforms():
    nl = navis.example_neurons(kind='skeleton')

    # Convert example neurons from voxels (raw hemibrain space) to microns
    tr = navis.xform_brain(nl, source='JRCFIB2018Fraw', target='JRCFIB2018Fum')

    assert isinstance(tr, navis.NeuronList)
    assert len(tr) == len(nl)
    assert all(nl.n_nodes == tr.n_nodes)


def test_meshneuron_transforms():
    nl = navis.example_neurons(kind='mesh')

    # Convert example neurons from voxels (raw hemibrain space) to microns
    tr = navis.xform_brain(nl, source='JRCFIB2018Fraw', target='JRCFIB2018Fum')

    assert isinstance(tr, navis.NeuronList)
    assert len(tr) == len(nl)
    assert all(nl.n_vertices == tr.n_vertices)


def test_dotprops_transforms():
    nl = navis.example_neurons(kind='skeleton')
    nl = navis.make_dotprops(nl, k=5)

    # Convert example neurons from voxels (raw hemibrain space) to microns
    tr = navis.xform_brain(nl, source='JRCFIB2018Fraw', target='JRCFIB2018Fum')

    assert isinstance(tr, navis.NeuronList)
    assert len(tr) == len(nl)
    assert all(nl.n_points == tr.n_points)


def test_volume_transforms():
    vol = navis.example_volume('LH')

    # Convert example neurons from voxels (raw hemibrain space) to microns
    tr = navis.xform_brain(vol, source='JRCFIB2018Fraw', target='JRCFIB2018Fum')

    assert isinstance(vol, navis.Volume)
    assert vol.vertices.shape[0] == tr.vertices.shape[0]


def test_skeleton_mirror():
    nl = navis.example_neurons(kind='skeleton')

    # Mirror along midline (produces garbage results but oh well)
    tr = navis.mirror_brain(nl, template='JRCFIB2018Fraw', warp=False)

    assert isinstance(tr, navis.NeuronList)
    assert len(tr) == len(nl)
    assert all(nl.n_nodes == tr.n_nodes)


def test_meshneuron_mirror():
    nl = navis.example_neurons(kind='mesh')

    # Mirror along midline (produces garbage results but oh well)
    tr = navis.mirror_brain(nl, template='JRCFIB2018Fraw', warp=False)

    assert isinstance(tr, navis.NeuronList)
    assert len(tr) == len(nl)
    assert all(nl.n_vertices == tr.n_vertices)


def test_dotprops_mirror():
    nl = navis.example_neurons(kind='skeleton')
    nl = navis.make_dotprops(nl, k=5)

    # Mirror along midline (produces garbage results but oh well)
    tr = navis.mirror_brain(nl, template='JRCFIB2018Fraw', warp=False)

    assert isinstance(tr, navis.NeuronList)
    assert len(tr) == len(nl)
    assert all(nl.n_points == tr.n_points)


def test_volume_mirror():
    vol = navis.example_volume('LH')

    # Mirror along midline (produces garbage results but oh well)
    tr = navis.mirror_brain(vol, template='JRCFIB2018Fraw', warp=False)

    assert isinstance(vol, navis.Volume)
    assert vol.vertices.shape[0] == tr.vertices.shape[0]



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Module for decorators.

Important: defer importing from other navis modules to avoid circular imports!
Even if this doesn't cause immediate issues it might well break pickling
functions (e.g. for multiprocessing).

"""

import re
import os
import inspect
import warnings

import numpy as np
import pandas as pd

from functools import wraps
from textwrap import dedent, indent

from typing import Optional, Union, List, Iterable, Dict, Tuple, Any
from typing_extensions import Literal

from .iterables import is_iterable, make_iterable


def map_neuronlist(
    desc: str = "",
    can_zip: List[Union[str, int]] = [],
    must_zip: List[Union[str, int]] = [],
    allow_parallel: bool = False,
):
    """Decorate function to run on all neurons in the NeuronList.

    This also updates the docstring.

    Parameters
    ----------
    desc :           str
                     Descriptor to show in the progress bar if run over multiple
                     neurons.
    can_zip/
    must_zip :       list
                     Names of keyword arguments that need to be zipped together
                     with the neurons in the neuronlist. For example:

                       some_function(NeuronList([n1, n2, n3]), [p1, p2, p3])

                     Should be executed as:

                       some_function(n1, p1)
                       some_function(n2, p2)
                       some_function(n3, p3)

                     `can_zip` will be zipped only if the length matches the
                     length of the neuronlist. If a `can_zip` argument has only
                     one value it will be re-used for all neurons.

                     `must_zip` arguments have to have one value for each of the
                     neurons.

                     Single `None` values are always just passed through.

                     Note that for this to consistently work the parameters in
                     question have to be keyword-only (*).
    allow_parallel : bool
                     If True and the function is called with `parallel=True`,
                     will use multiple cores to process the neuronlist. Number
                     of cores a can be set using `n_cores` keyword argument.

    """

    # TODO:
    # - make can_zip/must_zip work with positional-only argumens to, i.e. let
    #   it work with integers instead of strings
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            from .. import core

            # Get the function's signature
            sig = inspect.signature(function)

            try:
                fnname = function.__name__
            except BaseException:
                fnname = str(function)

            parallel = kwargs.pop("parallel", False)
            if parallel and not allow_parallel:
                raise ValueError(
                    f"Function {fnname} does not support parallel processing."
                )

            # First, we need to extract the neuronlist
            if args:
                # If there are positional arguments, the first one is
                # the input neuron(s)
                nl = args[0]
                nl_key = "__args"
            else:
                # If not, we need to look for the name of the first argument
                # in the signature
                nl_key = list(sig.parameters.keys())[0]
                nl = kwargs.get(nl_key, None)

            # Complain if we did not get what we expected
            if isinstance(nl, type(None)):
                raise ValueError(
                    "Unable to identify the neurons for call"
                    f"{fnname}:\n {args}\n {kwargs}"
                )

            # If we have a neuronlist
            if isinstance(nl, core.NeuronList):
                # Pop the neurons from kwargs or args so we don't pass the
                # neurons twice
                if nl_key == "__args":
                    args = args[1:]
                else:
                    _ = kwargs.pop(nl_key)

                # Check "can zip" arguments
                for p in can_zip:
                    # Skip if not present or is None
                    if p not in kwargs or isinstance(kwargs[p], type(None)):
                        continue

                    if is_iterable(kwargs[p]):
                        # If iterable but length does not match: complain
                        le = len(kwargs[p])
                        if le != len(nl):
                            raise ValueError(
                                f"Got {le} values of `{p}` for {len(nl)} neurons."
                            )

                # Parse "must zip" arguments
                for p in must_zip:
                    # Skip if not present or is None
                    if p not in kwargs or isinstance(kwargs[p], type(None)):
                        continue

                    values = make_iterable(kwargs[p])
                    if len(values) != len(nl):
                        raise ValueError(
                            f"Got {len(values)} values of `{p}` for {len(nl)} neurons."
                        )

                # If we use parallel processing it makes sense to modify neurons
                # "inplace" since they will be copied into the child processes
                # anyway and that way we can avoid making an additional copy
                if "inplace" in kwargs:
                    # First check keyword arguments
                    inplace = kwargs["inplace"]
                elif "inplace" in sig.parameters:
                    # Next check signatures default
                    inplace = sig.parameters["inplace"].default
                else:
                    # All things failing assume it's not inplace
                    inplace = False

                if parallel and "inplace" in sig.parameters:
                    kwargs["inplace"] = True

                # Prepare processor
                n_cores = kwargs.pop("n_cores", os.cpu_count() // 2)
                chunksize = kwargs.pop("chunksize", 1)
                excl = list(kwargs.keys()) + list(range(1, len(args) + 1))
                proc = core.NeuronProcessor(
                    nl,
                    function,
                    parallel=parallel,
                    desc=desc,
                    warn_inplace=False,
                    progress=kwargs.pop("progress", True),
                    omit_failures=kwargs.pop("omit_failures", False),
                    chunksize=chunksize,
                    exclude_zip=excl,
                    n_cores=n_cores,
                )
                # Apply function
                res = proc(nl, *args, **kwargs)

                # When using parallel processing, the neurons will not actually
                # have been modified inplace - in that case we will simply
                # replace the neurons in `nl`
                if inplace:
                    nl.neurons = res.neurons
                else:
                    nl = res

                return nl
            else:
                # If single neuron just pass through
                return function(*args, **kwargs)

        # Update the docstring
        wrapper = map_neuronlist_update_docstring(wrapper, allow_parallel)

        return wrapper

    return decorator


def map_neuronlist_df(
    desc: str = "",
    id_col: str = "neuron",
    reset_index: bool = True,
    allow_parallel: bool = False,
):
    """Decorate function to run on all neurons in the NeuronList.

    This version of the decorator is meant for functions that return a
    DataFrame. This decorator will add a `neuron` column with the respective
    neuron's ID and will then concatenate the dataframes.

    Parameters
    ----------
    desc :           str
                     Descriptor to show in the progress bar if run over multiple
                     neurons.
    id_col :         str
                     Name of the ID column to be added to the results dataframe.
    reset_index :    bool
                     Whether to reset the index of the dataframe after
                     concatenating.
    allow_parallel : bool
                     If True and the function is called with `parallel=True`,
                     will use multiple cores to process the neuronlist. Number
                     of cores a can be set using `n_cores` keyword argument.

    """

    # TODO:
    # - make can_zip/must_zip work with positional-only argumens to, i.e. let
    #   it work with integers instead of strings
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            # Lazy import to avoid issues with circular imports and pickling
            from .. import core

            # Get the function's signature
            sig = inspect.signature(function)

            try:
                fnname = function.__name__
            except BaseException:
                fnname = str(function)

            parallel = kwargs.pop("parallel", False)
            if parallel and not allow_parallel:
                raise ValueError(
                    f"Function {fnname} does not allow parallel processing."
                )

            # First, we need to extract the neuronlist
            if args:
                # If there are positional arguments, the first one is
                # the input neuron(s)
                nl = args[0]
                nl_key = "__args"
            else:
                # If not, we need to look for the name of the first argument
                # in the signature
                nl_key = list(sig.parameters.keys())[0]
                nl = kwargs.get(nl_key, None)

            # Complain if we did not get what we expected
            if isinstance(nl, type(None)):
                raise ValueError(
                    "Unable to identify the neurons for call"
                    f"{fnname}:\n {args}\n {kwargs}"
                )

            # If we have a neuronlist
            if isinstance(nl, core.NeuronList):
                # Pop the neurons from kwargs or args so we don't pass the
                # neurons twice
                if nl_key == "__args":
                    args = args[1:]
                else:
                    _ = kwargs.pop(nl_key)

                # Prepare processor
                n_cores = kwargs.pop("n_cores", os.cpu_count() // 2)
                chunksize = kwargs.pop("chunksize", 1)
                excl = list(kwargs.keys()) + list(range(1, len(args) + 1))
                proc = core.NeuronProcessor(
                    nl,
                    function,
                    parallel=parallel,
                    desc=desc,
                    warn_inplace=False,
                    progress=kwargs.pop("progress", True),
                    omit_failures=kwargs.pop("omit_failures", False),
                    chunksize=chunksize,
                    exclude_zip=excl,
                    n_cores=n_cores,
                )
                # Apply function
                res = proc(nl, *args, **kwargs)

                for n, df in zip(nl, res):
                    df.insert(0, column=id_col, value=n.id)

                df = pd.concat(res, axis=0)

                if reset_index:
                    df = df.reset_index(drop=True)

            else:
                # If single neuron just pass through
                df = function(*args, **kwargs)
                # df.insert(0, column=id_col, value=nl.id)

            return df

        # Update the docstring
        wrapper = map_neuronlist_update_docstring(wrapper, allow_parallel)

        return wrapper

    return decorator


def map_neuronlist_update_docstring(func, allow_parallel):
    """Add additional parameters to docstring of function."""
    # Parse docstring
    lines = func.__doc__.split("\n")

    # Find a line with a parameter
    pline = [l for l in lines if " : " in l][0]
    # Get the leading whitespaces
    wspaces = " " * re.search("( *)", pline).end(1)
    # Get the offset for type and description
    offset = re.search("( *: *)", pline).end(1) - len(wspaces)

    # Find index of the last parameters (assuming there is a single empty
    # line between Returns and the last parameter)
    try:
        lastp = [
            i
            for i, line in enumerate(lines[:-1])
            if "Returns" in line and "----" in lines[i + 1]
        ][0] - 1
    except IndexError:
        warnings.warn(f'Could not find "Returns" in docstring for function {func}')
        return func

    msg = ""
    if allow_parallel:
        msg += dedent(f"""\
        parallel :{" " * (offset - 10)}bool
                  {" " * (offset - 10)}If True and input is NeuronList, use parallel
                  {" " * (offset - 10)}processing. Requires `pathos`.
        n_cores : {" " * (offset - 10)}int, optional
                  {" " * (offset - 10)}Numbers of cores to use if `parallel=True`.
                  {" " * (offset - 10)}Defaults to half the available cores.
        """)

    msg += dedent(f"""\
    progress :{" " * (offset - 10)}bool
              {" " * (offset - 10)}Whether to show a progress bar. Overruled by
              {" " * (offset - 10)}`navis.set_pbars`.
    omit_failures :{" " * (offset - 15)}bool
                   {" " * (offset - 15)}If True will omit failures instead of raising
                   {" " * (offset - 15)}an exception. Ignored if input is single neuron.
    """)

    # Insert new docstring
    lines.insert(lastp, indent(msg, wspaces))

    # Update docstring
    func.__doc__ = "\n".join(lines)

    return func


def lock_neuron(function):
    """Lock neuron while function is executed.

    This makes sure that temporary attributes aren't re-calculated as changes
    are being made.

    """

    @wraps(function)
    def wrapper(*args, **kwargs):
        # Lazy import to avoid issues with circular imports and pickling
        from .. import core

        # Lock if first argument is a neuron
        if isinstance(args[0], core.BaseNeuron):
            args[0]._lock = getattr(args[0], "_lock", 0) + 1
        try:
            # Execute function
            res = function(*args, **kwargs)
        except BaseException:
            raise
        finally:
            # Unlock neuron
            if isinstance(args[0], core.BaseNeuron):
                args[0]._lock -= 1
        # Return result
        return res

    return wrapper


def meshneuron_skeleton(
    method: Union[
        Literal["subset"],
        Literal["split"],
        Literal["node_properties"],
        Literal["node_to_vertex"],
        Literal["pass_through"],
    ],
    include_connectors: bool = False,
    copy_properties: list = [],
    disallowed_kwargs: dict = {},
    node_props: list = [],
    reroot_soma: bool = False,
    heal: bool = False,
):
    """Decorate function such that MeshNeurons are automatically skeletonized,
    the function is run on the skeleton and changes are propagated
    back to the meshe.

    Parameters
    ----------
    method :    str
                What to do with the results:
                  - 'subset': subset MeshNeuron to what's left of the skeleton
                  - 'split': split MeshNeuron following the skeleton's splits
                  - 'node_to_vertex': map the returned node ID to the vertex IDs
                  - 'node_properties' map node properties to vertices (requires
                    `node_props` parameter)
                  - 'pass_through' simply passes through the return value
    include_connectors : bool
                If True, will try to make sure that if the MeshNeuron has
                connectors, they will be carried over to the skeleton.
    copy_properties : list
                Any additional properties that need to be copied from the
                skeleton to the mesh.
    disallowed_kwargs : dict
                Keyword arguments (name + value) that are not permitted when
                input is MeshNeuron.
    node_props : list
                For method 'node_properties'. String must be column names in
                node table of skeleton.
    reroot_soma :  bool
                If True and neuron has a soma (.soma_pos), will reroot to
                that soma.
    heal :      bool
                Whether or not to heal the skeleton if the mesh is fragmented.

    """
    assert isinstance(copy_properties, list)
    assert isinstance(disallowed_kwargs, dict)
    assert isinstance(node_props, list)

    allowed_methods = (
        "subset",
        "node_to_vertex",
        "split",
        "node_properties",
        "pass_through",
    )
    if method not in allowed_methods:
        raise ValueError(f'Unknown method "{method}"')

    if method == "node_properties" and not node_props:
        raise ValueError('Must provide `node_props` for method "node_properties"')

    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            # Get the function's signature
            sig = inspect.signature(function)

            try:
                fnname = function.__name__
            except BaseException:
                fnname = str(function)

            # First, we need to extract the neuron from args and kwargs
            if args:
                # If there are positional arguments, the first one is assumed to
                # be the input neuron
                x = args[0]
                args = args[1:]
                x_key = "__args"
            else:
                # If not, we need to look for the name of the first argument
                # in the signature
                x_key = list(sig.parameters.keys())[0]
                x = kwargs.pop(x_key, None)

            # Complain if we did not get what we expected
            if isinstance(x, type(None)):
                raise ValueError(
                    "Unable to identify the neurons for call"
                    f"{fnname}:\n {args}\n {kwargs}"
                )

            # If input not a MeshNeuron, just pass through
            # Note delayed import to avoid circular imports and IMPORTANTLY
            # funky interactions with pickle/dill
            from .. import core

            if not isinstance(x, core.MeshNeuron):
                return function(x, *args, **kwargs)

            # Check for disallowed kwargs
            for k, v in disallowed_kwargs.items():
                if k in kwargs and kwargs[k] == v:
                    raise ValueError(
                        f"{k}={v} is not allowed when input is MeshNeuron(s)."
                    )

            # See if this is meant to be done inplace
            if "inplace" in kwargs:
                # First check keyword arguments
                inplace = kwargs["inplace"]
            elif "inplace" in sig.parameters:
                # Next check signatures default
                inplace = sig.parameters["inplace"].default
            else:
                # All things failing assume it's not inplace
                inplace = False

            # Now skeletonize
            sk = x.skeleton

            # Delayed import to avoid circular imports
            # Note that this HAS to be in the inner function otherwise
            # we get a weird error when pickling for parallel processing
            from .. import morpho

            if heal:
                sk = morpho.heal_skeleton(sk, method="LEAFS")

            if reroot_soma and sk.has_soma:
                sk = sk.reroot(sk.soma)

            if include_connectors and x.has_connectors and not sk.has_connectors:
                sk._connectors = x.connectors.copy()
                sk._connectors["node_id"] = sk.snap(
                    sk.connectors[["x", "y", "z"]].values
                )[0]

            # Apply function
            res = function(sk, *args, **kwargs)

            if method == "subset":
                # See which vertices we need to keep
                keep = np.isin(sk.vertex_map, res.nodes.node_id.values)

                x = morpho.subset_neuron(x, keep, inplace=inplace)

                for p in copy_properties:
                    setattr(x, p, getattr(sk, p, None))
            elif method == "split":
                meshes = []
                for n in res:
                    # See which vertices we need to keep
                    keep = np.isin(sk.vertex_map, n.nodes.node_id.values)

                    meshes.append(morpho.subset_neuron(x, keep, inplace=False))

                    for p in copy_properties:
                        setattr(meshes[-1], p, getattr(n, p, None))
                x = core.NeuronList(meshes)
            elif method == "node_to_vertex":
                x = np.where(sk.vertex_map == res)[0]
            elif method == "node_properties":
                for p in node_props:
                    node_map = sk.nodes.set_index("node_id")[p].to_dict()
                    vertex_props = np.array([node_map[n] for n in sk.vertex_map])
                    setattr(x, p, vertex_props)
            elif method == "pass_through":
                return res

            return x

        return wrapper

    return decorator



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import skeletor as sk
import trimesh as tm
import networkx as nx

from numbers import Number
from scipy.ndimage import gaussian_filter
from typing import Union, Optional

from .. import core, config, utils, morpho, graph

logger = config.get_logger(__name__)


@utils.map_neuronlist(desc='Skeletonizing', allow_parallel=True)
def points2skeleton(x: Union['core.Dotprops', np.ndarray],
                    k: int = 10,
                    max_dist: Optional[float] = None):
    """Turn points into skeleton.

    This function works by:
     1. Compute the `k` nearest neighbors for each point
     2. Generate a graph from the nearest-neighbor edges
     3. Extract a minimum-spanning tree (MST) from the graph
     4. Process the MST into a skeleton

    Parameters
    ----------
    x :         (N, 3) array | Dotprops
                Points to skeletonize.
    k :         int
                Number of nearest neighbors to consider. Too low values of `k`
                can lead to disconnected skeletons.
    max_dist :  float, optional
                Edges longer than this will be ignored. This can lead to a
                fragmented (i.e. multi-root) skeleton!

    Returns
    -------
    skeleton :  navis.TreeNeuron

    Examples
    --------
    >>> import navis
    >>> # Get a mesh neuron
    >>> n = navis.example_neurons(1)
    >>> # Get the points
    >>> pts = n.nodes[['x', 'y', 'z']].values
    >>> # Convert points back into skeleton
    >>> sk = navis.conversion.points2skeleton(pts)

    """
    utils.eval_param(x, name='x', allowed_types=(core.Dotprops, np.ndarray))

    if isinstance(x, core.Dotprops):
        pts = x.points
    else:
        if (x.ndim != 2) and (x.shape[1] != 3):
            raise ValueError(f'Points must be shape (N, 3), got {x.shape}')
        pts = x

    # Get the list of nearest neighbours
    tree = core.dotprop.KDTree(pts)

    defaults = {}
    if max_dist is not None:
        # We have to avoid passing `None` because scipy's KDTree does not like
        # that (pykdtree does not care)
        defaults['distance_upper_bound'] = max_dist
    dists, NN = tree.query(pts, k=k + 1, **defaults)

    # Drop self-hits
    dists, NN = dists[:, 1:], NN[:, 1:]

    # Turn into edges
    edges = []
    ix1 = np.arange(len(dists))
    for i in range(k):
        ix2 = NN[:, i]
        le = dists[:, i]
        # If a max dist was set we have to remove NN that have dist np.inf
        if max_dist is None:
            edges += list(zip(ix1, ix2, le))
        else:
            not_inf = le != np.inf
            edges += list(zip(ix1[not_inf], ix2[not_inf], le[not_inf]))

    # Generate graph
    G = nx.Graph()
    G.add_nodes_from(ix1)
    G.add_weighted_edges_from(edges)

    # Extract minimum spanning tree
    G_mst = nx.minimum_spanning_tree(G)

    # Add the coordinates as node properties
    nx.set_node_attributes(G_mst, dict(zip(G.nodes, pts[:, 0])), name='x')
    nx.set_node_attributes(G_mst, dict(zip(G.nodes, pts[:, 1])), name='y')
    nx.set_node_attributes(G_mst, dict(zip(G.nodes, pts[:, 2])), name='z')

    return graph.nx2neuron(G_mst)


@utils.map_neuronlist(desc='Skeletonizing', allow_parallel=True)
def mesh2skeleton(x: 'core.MeshNeuron',
                  method: str = 'wavefront',
                  fix_mesh: bool = False,
                  shave: bool = True,
                  heal: bool = False,
                  connectors: bool = False,
                  inv_dist: Union[int, float] = None,
                  **kwargs):
    """Turn mesh neuron into skeleton.

    This function is a thin-wrapper for `skeletor`. It uses sensible defaults
    for neurons but if you want to fine-tune your skeletons you should look
    into using `skeletor` directly.

    Parameters
    ----------
    x :         MeshNeuron | trimesh.Trimesh
                Mesh(es) to skeletonize. Note that the quality of the results
                very much depends on the mesh, so it might be worth doing some
                pre-processing (see below).
    method :    'wavefront' | 'teasar'
                Method to use for skeletonization:
                 - "wavefront": fast but noisier, skeletons will be ~centered
                   within the neuron
                 - "teasar": slower but smoother, skeletons follow the
                   surface of the mesh, requires the `inv_dist` parameter to be
                   set
                "wavefront" also produces radii, "teasar" doesn't.
    fix_mesh :  bool
                Whether to try to fix some common issues in the mesh before
                skeletonization. Note that this might compromise the
                vertex-to-node-ID mapping.
    shave :     bool
                Whether to "shave" the resulting skeleton to reduce bristles
                on the backbone.
    heal :      bool | "LEAFS" | "ALL"
                Whether to heal the resulting skeleton if it is fragmented.
                For more control over the stitching set `heal=False` and use
                [`navis.heal_skeleton`][] directly. Note that this
                can be fairly costly if the mesh as many tiny fragments.
    connectors : bool
                Whether to carry over existing connector tables. This will
                attach connectors by first snapping them to the closest mesh
                vertex and then to the skeleton node corresponding to that
                vertex.
    inv_dist :  int | float | str
                Only required for method "teasar": invalidation distance for
                the traversal. Smaller `inv_dist` captures smaller features
                but is slower and more noisy, and vice versa. A good starting
                value is around 2-5 microns. Can be a unit string - e.g.
                "5 microns" - if your neuron has its units set.
    **kwargs
                Additional keyword arguments are passed through to the respective
                function in `skeletor` - i.e. `by_wavefront` or `by_teasar`.

    Returns
    -------
    skeleton :  navis.TreeNeuron
                Has a `.vertex_map` attribute that maps each vertex in the
                input mesh to a skeleton node ID.

    See Also
    --------
    [`navis.drop_fluff`][]
                Use this if your mesh has lots of tiny free floating bits to
                reduce noise and speed up skeletonization.

    Examples
    --------
    >>> import navis
    >>> # Get a mesh neuron
    >>> n = navis.example_neurons(1, kind='mesh')
    >>> # Convert to skeleton
    >>> sk = navis.conversion.mesh2skeleton(n)
    >>> # Mesh vertex indices to node IDs map
    >>> sk.vertex_map                                           # doctest: +SKIP
    array([938, 990, 990, ...,  39, 234, 234])

    """
    utils.eval_param(x, name='x', allowed_types=(core.MeshNeuron, tm.Trimesh))
    utils.eval_param(method, name='method', allowed_values=('wavefront', 'teasar'))

    if method == 'teasar' and inv_dist is None:
        raise ValueError('Must set `inv_dist` parameter when using method '
                         '"teasar". A good starting value is around 2-5 microns.')

    props = {'soma': None}
    if isinstance(x, core.MeshNeuron):
        props.update({'id': x.id, 'name': x.name, 'units': x.units})
        if x.has_soma_pos:
            props['soma_pos'] = x.soma_pos

        if not isinstance(inv_dist, type(None)):
            inv_dist = x.map_units(inv_dist)

        mesh = x.trimesh
    else:
        mesh = x
        x = core.MeshNeuron(x)

    if fix_mesh:
        mesh = sk.pre.fix_mesh(mesh, remove_disconnected=False)

    kwargs['progress'] = False
    if method == 'wavefront':
        skeleton = sk.skeletonize.by_wavefront(mesh, **kwargs)
    elif method == 'teasar':
        skeleton = sk.skeletonize.by_teasar(x, inv_dist=inv_dist, **kwargs)

    props['vertex_map'] = skeleton.mesh_map

    s = core.TreeNeuron(skeleton.swc, **props)

    if s.has_soma:
        s.reroot(s.soma, inplace=True)

    if heal:
        _ = morpho.heal_skeleton(s, inplace=True, method='ALL')

    if shave:
        # Find single node bristles
        leafs = s.leafs.node_id.values

        # Make sure we keep the soma
        if s.has_soma:
            leafs = leafs[~np.isin(leafs, s.soma)]

        bp = s.branch_points.node_id.values
        bristles = s.nodes[s.nodes.node_id.isin(leafs)
                           & s.nodes.parent_id.isin(bp)]

        # Subset neuron
        keep = s.nodes[~s.nodes.node_id.isin(bristles.node_id)].node_id.values
        s = morpho.subset_neuron(s, keep, inplace=True)

        # Fix vertex map
        for b, p in zip(bristles.node_id.values, bristles.parent_id.values):
            s.vertex_map[s.vertex_map == b] = p

    # In particular with method wavefront, some nodes (mostly leafs) can have
    # a radius of 0. We will fix this here by giving them 1/2 the radius of
    # their parent nodes'
    to_fix = (s.nodes.radius == 0) & (s.nodes.parent_id >= 0)
    if any(to_fix):
        radii = s.nodes.set_index('node_id').radius
        new_radii = radii.loc[s.nodes.loc[to_fix].parent_id].values / 2
        s.nodes.loc[to_fix, 'radius'] = new_radii

    # Last but not least: map connectors
    if connectors and x.has_connectors:
        cn_table = x.connectors.copy()

        # A connector/id column is currently required for skeletons but not
        # meshes
        if not any(np.isin(('id', 'connector_id'), cn_table.columns)):
            cn_table.insert(0, 'connector_id', np.arange(len(cn_table)))

        cn_table['node_id'] = x.snap(cn_table[['x', 'y', 'z']].values)[0]
        node_map = dict(zip(np.arange(len(s.vertex_map)), s.vertex_map))
        cn_table['node_id'] = cn_table.node_id.map(node_map)
        s.connectors = cn_table

    return s


def _make_voxels(x: 'core.BaseNeuron',
                 pitch: Union[list, tuple, float],
                 strip: bool = False):
    """Turn neuron into voxels.

    Parameters
    ----------
    x :         TreeNeuron | MeshNeuron | Dotprops
                Neuron(s) to voxelize. Uses the neurons' nodes, vertices and
                points, respectively.
    pitch :     float | iterable thereof
                Side length(s) of voxels. Can be isometric (float) or an
                iterable of dimensions in (x, y, z).
    strip :     bool
                Whether to strip empty leading voxels.

    Returns
    -------
    voxels :    numpy array
                Array of voxel indices.
    offset :    (3, ) numpy array
                Offset for voxels. Will be (0, 0, 0) if `strip=False`.

    See Also
    --------
    [`navis.neuron2voxelgrid`][]
                Use this function to create a voxel neuron covering a specific
                volume. Useful e.g. when wanting to directly compare two
                neurons.

    """
    utils.eval_param(x, name='x', allowed_types=(core.BaseNeuron, tm.Trimesh))

    if not utils.is_iterable(pitch):
        if not isinstance(pitch, Number):
            raise TypeError('Expected `pitch` to be a number (or list thereof)'
                            f', got {type(pitch)}')
        pitch = [pitch] * 3
    elif len(pitch) != 3:
        raise ValueError('`pitch` must be single number or a list of three')

    if isinstance(x, core.TreeNeuron):
        pts = x.nodes[['x', 'y', 'z']].values
    elif isinstance(x, core.Dotprops):
        pts = x.points
    elif isinstance(x, (core.MeshNeuron, tm.Trimesh)):
        pts = np.array(x.vertices)

    # Convert points to voxel indices
    ix = (pts / pitch).round().astype(int)

    # Make neuron
    if strip:
        offset = ix.min(axis=0) * pitch
        ix = ix - ix.min()
    else:
        offset = (0, 0, 0)

    return ix, offset


@utils.map_neuronlist(desc='Voxelizing', allow_parallel=True)
def neuron2voxels(x: 'core.BaseNeuron',
                  pitch: Union[list, tuple, float],
                  bounds: Optional[list] = None,
                  counts: bool = False,
                  vectors: bool = False,
                  alphas: bool = False,
                  smooth: int = 0) -> 'core.VoxelNeuron':
    """Turn neuron into voxels.

    Parameters
    ----------
    x :             TreeNeuron | MeshNeuron | Dotprops
                    Neuron(s) to voxelize. Uses the neurons' nodes, vertices and
                    points, respectively.
    pitch :         float | iterable thereof
                    Side length(s) voxels. Can be isometric (float) or an
                    iterable of dimensions in (x, y, z).
    bounds :        (3, 2)  or (2, 3) array, optional
                    Boundaries [in units of `x`] for the voxel grid. If not
                    provided, will use `x.bbox`.
    counts :        bool
                    If True, voxel grid will have point counts for values
                    instead of just True/False.
    vectors :       bool
                    If True, will also attach a vector field as `.vectors`
                    property.
    alphas :        bool
                    If True, will also return a grid with alpha values as
                    `.alpha` property.
    smooth :        int
                    If non-zero, will apply a Gaussian filter with `smooth`
                    as `sigma`.

    Returns
    -------
    VoxelNeuron
                    Has the voxel grid as `.grid` and (optionally) `.vectors`
                    and `.alphas` properties. `.grid` data type depends
                    on settings:
                     - default = bool (i.e. True/False)
                     - if `counts=True` = integer
                     - if `smooth=True` = float
                    Empty voxels will have vector (0, 0, 0) and alpha 0. Also
                    note that data tables (e.g. `connectors`) are not carried
                    over from the input neuron.

    Examples
    --------
    >>> import navis
    >>> # Get a skeleton
    >>> n = navis.example_neurons(1)
    >>> # Convert to voxel neuron
    >>> vx = navis.conversion.neuron2voxels(n, pitch='5 microns')

    """
    if not utils.is_iterable(pitch):
        # Map units (non-str are just passed through)
        pitch = x.map_units(pitch, on_error='raise')
        if not isinstance(pitch, Number):
            raise TypeError('Expected `pitch` to be a number (or list thereof)'
                            f', got {type(pitch)}')
        pitch = [pitch] * 3
    elif len(pitch) != 3:
        raise ValueError('`pitch` must be single number or a list of three')
    else:
        pitch = np.array([x.map_units(p, on_error='raise') for p in pitch])

    # Convert to voxel indices
    ix, _ = _make_voxels(x=x, pitch=pitch, strip=False)

    if isinstance(bounds, type(None)):
        bounds = x.bbox
    else:
        bounds = np.asarray(bounds)

    if bounds.shape == (2, 3):
        bounds = bounds.T

    # Shape of grid
    dim = np.ceil(bounds[:, 1] / pitch) - np.floor(bounds[:, 0] / pitch)
    shape = np.ceil(dim).astype(int) + 1

    # Get unique voxels
    if not counts:
        vxl = np.unique(ix, axis=0)
    else:
        vxl, cnt = np.unique(ix, axis=0, return_counts=True)

    # Substract lower bounds
    offset = (bounds[:, 0] / pitch)
    vxl = vxl - offset.round().astype(int)
    ix = ix - offset.round().astype(int)

    # Drop voxels outside the defined bounds
    vxl = vxl[vxl.min(axis=1) >= 0]
    vxl = vxl[np.all(vxl < shape, axis=1)]

    # Generate grid
    grid = np.zeros(shape=shape, dtype=bool)

    # Populate grid
    if not counts:
        grid[vxl[:, 0], vxl[:, 1], vxl[:, 2]] = True
    else:
        grid = grid.astype(int)
        grid[vxl[:, 0], vxl[:, 1], vxl[:, 2]] = cnt

    # Apply Gaussian filter
    if smooth:
        grid = gaussian_filter(grid.astype(np.float32), sigma=smooth)

    # Generate neuron
    units = [f'{p * u} {x.units.units}' for p, u in zip(utils.make_iterable(pitch),
                                                        x.units_xyz.magnitude)]
    offset = offset * pitch * x.units_xyz.magnitude
    n = core.VoxelNeuron(grid, id=x.id, name=x.name, units=units, offset=offset)

    # If no vectors required, we can just return now
    if not vectors and not alphas:
        return n

    if isinstance(x, core.TreeNeuron):
        pts = x.nodes[['x', 'y', 'z']].values
    elif isinstance(x, core.Dotprops):
        pts = x.points
    elif isinstance(x, core.MeshNeuron):
        pts = np.array(x.vertices)

    # Generate an empty vector field
    vects = np.zeros((grid.shape[0], grid.shape[1], grid.shape[2], 3),
                     dtype=np.float32)
    alph = np.zeros(grid.shape, dtype=np.float32)

    # Get unique voxels
    uni, inv = np.unique(ix, axis=0, return_inverse=True)

    # Go over each voxel
    for i in range(len(uni)):
        # Get points in this voxel
        pt = pts[inv == i]

        # Reshape
        pt = pt.reshape(1, -1, 3)

        # Generate centers for each cloud of k nearest neighbors
        centers = np.mean(pt, axis=1)

        # Generate vector from center
        cpt = pt - centers.reshape((pt.shape[0], 1, 3))

        # Get inertia (N, 3, 3)
        inertia = cpt.transpose((0, 2, 1)) @ cpt

        # Extract vector and alpha
        u, s, vh = np.linalg.svd(inertia)
        vect = vh[:, 0, :]

        # No alpha if only one point
        if pt.shape[1] > 1:
            alpha = (s[:, 0] - s[:, 1]) / np.sum(s, axis=1)
        else:
            alpha = [0]

        vects[uni[i][0], uni[i][1], uni[i][2]] = vect.flatten()
        alph[uni[i][0], uni[i][1], uni[i][2]] = alpha[0]

    if vectors:
        n.vectors = vects
    if alpha:
        n.alphas = alpha

    return n


@utils.map_neuronlist(desc='Converting', allow_parallel=True)
def tree2meshneuron(x: 'core.TreeNeuron',
                    tube_points: int = 8,
                    radius_scale_factor: float = 1,
                    use_normals: bool = True,
                    warn_missing_radii: bool = True
                    ) -> 'core.MeshNeuron':
    """Convert TreeNeuron to MeshNeuron.

    Uses the `radius` to convert skeleton to 3D tube mesh. Missing radii are
    treated as zeros.

    Parameters
    ----------
    x :             TreeNeuron | NeuronList
                    Neuron to convert.
    tube_points :   int
                    Number of points making up the circle of the cross-section
                    of the tube.
    radius_scale_factor : float
                    Factor to scale radii by.
    use_normals :   bool
                    If True will rotate tube along its curvature.
    warn_missing_radii : bool
                    Whether to warn if radii are missing or <= 0.

    Returns
    -------
    TreeNeuron
                    Data tables (e.g. `connectors`) are not carried over from
                    the input neuron.

    Examples
    --------
    >>> import navis
    >>> # Get a skeleton
    >>> n = navis.example_neurons(1)
    >>> # Convert to mesh neuron
    >>> m = navis.conversion.tree2meshneuron(n)

    """
    # Delay to avoid circular imports
    from ..plotting.plot_utils import make_tube

    if not isinstance(x, core.TreeNeuron):
        raise TypeError(f'Expected TreeNeuron, got "{type(x)}"')

    # Map segments of node IDs to segments of node indices
    id2ix = dict(zip(x.nodes.node_id, np.arange(len(x.nodes))))
    segments = [np.array([id2ix[n] for n in seg]) for seg in x.segments]

    # Note that we are treating missing radii as "0"
    radii_map = x.nodes.radius.fillna(0).values
    if warn_missing_radii and (radii_map <= 0).any():
        logger.warning('At least some radii are missing or <= 0. Mesh may look funny.')

    # Map radii onto segments
    radii = [radii_map[seg] * radius_scale_factor for seg in segments]
    co_map = x.nodes[['x', 'y', 'z']].values
    seg_points = [co_map[seg] for seg in segments]

    vertices, faces = make_tube(seg_points,
                                radii=radii,
                                tube_points=tube_points,
                                use_normals=use_normals)

    # Note: the `process=False` is necessary to not break correspondence
    # by e.g. merging duplicate vertices
    m = core.MeshNeuron({'vertices': vertices, 'faces': faces},
                        units=x.units, name=x.name, id=x.id, process=False)


    # For each vertex, track the original node: the first `tube_points` vertices
    # correspond to the first node of the first segment and so on.
    m.vertex_map = np.concatenate([np.repeat(seg, tube_points) for seg in segments])

    return m



"""Configuration for in-package doctests.

Should be importable (but not useful)
without development dependencies.

Where these files up and when they are deleted is documented
[here](https://pytest.org/en/stable/tmpdir.html#the-default-base-temporary-directory).
"""
from pathlib import Path

try:
    import pytest
    decorator = pytest.fixture(autouse=True)
except ModuleNotFoundError:
    def decorator(fn):
        return fn


@decorator
def add_tmp_dir(doctest_namespace, tmpdir):
    """Give all doctests access to a `tmp_dir` variable.

    `tmp_dir` is a `pathlib.Path` to a real directory
    in pytest's tmp directory which is automatically cleaned up
    in later invocations of `pytest`.
    """
    doctest_namespace["tmp_dir"] = Path(tmpdir)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2017 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along

"""Module contains functions to plot neurons in 2D and 3D."""

from .. import config, core, utils

import math
import random
import warnings

import numpy as np

from collections.abc import Iterable
from typing import Tuple, Optional, List, Dict


__all__ = ["tn_pairs_to_coords", "segments_to_coords", "fibonacci_sphere", "make_tube"]

logger = config.get_logger(__name__)


def tn_pairs_to_coords(
    x: core.TreeNeuron, modifier: Optional[Tuple[float, float, float]] = (1, 1, 1)
) -> np.ndarray:
    """Return pairs of child->parent node coordinates.

    Parameters
    ----------
    x :         TreeNeuron
                Must contain the nodes.
    modifier :  ints, optional
                Use to modify/invert x/y/z axes.

    Returns
    -------
    coords :    np.array
                `[[[x1, y1, z1], [x2, y2, z2]], [[x3, y3, y4], [x4, y4, z4]]]`

    """
    if not isinstance(modifier, np.ndarray):
        modifier = np.array(modifier)

    nodes = x.nodes[x.nodes.parent_id >= 0]

    tn_co = nodes.loc[:, ["x", "y", "z"]].values
    parent_co = (
        x.nodes.set_index("node_id").loc[nodes.parent_id.values, ["x", "y", "z"]].values
    )

    coords = np.append(tn_co, parent_co, axis=1)

    if any(modifier != 1):
        coords *= modifier

    return coords.reshape((coords.shape[0], 2, 3))


def segments_to_coords(
    x: core.TreeNeuron,
    modifier: Optional[Tuple[float, float, float]] = (1, 1, 1),
    node_colors: Optional[np.ndarray] = None,
) -> List[np.ndarray]:
    """Turn lists of node IDs into coordinates.

    Will use navis-fastcore if available.

    Parameters
    ----------
    x :             TreeNeuron
                    Must contain the nodes
    node_colors :   numpy.ndarray, optional
                    A color for each node in `x.nodes`. If provided, will
                    also return a list of colors sorted to match coordinates.
    modifier :      ints, optional
                    Use e.g. to modify/invert x/y/z axes.

    Returns
    -------
    coords :        list of tuples
                    [(x, y, z), (x, y, z), ... ]
    colors :        list of colors
                    If `node_colors` provided will return a copy of it sorted
                    to match `coords`.

    """
    colors = None

    if utils.fastcore is not None:
        if node_colors is None:
            coords = utils.fastcore.segment_coords(
                x.nodes.node_id.values,
                x.nodes.parent_id.values,
                x.nodes[["x", "y", "z"]].values,
            )
        else:
            coords, colors = utils.fastcore.segment_coords(
                x.nodes.node_id.values,
                x.nodes.parent_id.values,
                x.nodes[["x", "y", "z"]].values,
                node_colors=node_colors,
            )
    else:
        # Using a dictionary here is orders of manitude faster than .loc[]!
        locs: Dict[int, Tuple[float, float, float]]
        # Oddly, this is also the fastest way to generate the dictionary
        nodes = x.nodes
        locs = {
            i: (x, y, z)
            for i, x, y, z in zip(
                nodes.node_id.values, nodes.x.values, nodes.y.values, nodes.z.values
            )
        }  # type: ignore
        # locs = {r.node_id: (r.x, r.y, r.z) for r in x.nodes.itertuples()}  # type: ignore
        coords = [np.array([locs[tn] for tn in s]) for s in x.segments]

        if not isinstance(node_colors, type(None)):
            ilocs = dict(zip(x.nodes.node_id.values, np.arange(x.nodes.shape[0])))
            colors = [node_colors[[ilocs[tn] for tn in s]] for s in x.segments]

    modifier = np.asarray(modifier)
    if (modifier != 1).any():
        for seg in coords:
            np.multiply(seg, modifier, out=seg)

    if colors is not None:
        return coords, colors

    return coords


def fibonacci_sphere(samples: int = 1, randomize: bool = True) -> list:
    """Generate points on a sphere."""
    rnd = 1.0
    if randomize:
        rnd = random.random() * samples

    points = []
    offset = 2.0 / samples
    increment = math.pi * (3.0 - math.sqrt(5.0))

    for i in range(samples):
        y = ((i * offset) - 1) + (offset / 2)
        r = math.sqrt(1 - pow(y, 2))

        phi = ((i + rnd) % samples) * increment

        x = math.cos(phi) * r
        z = math.sin(phi) * r

        points.append([x, y, z])

    return np.array(points)


def make_tube(segments, radii=1.0, tube_points=8, use_normals=True):
    """Generate tube mesh (vertices + faces) from lines.

    This code was modified from the vispy library.

    Parameters
    ----------
    segments :      list
                    List of lists of x/y/z coordinates.
    radii :         float | list of floats
                    Either a single radius used for all nodes or list of lists of
                    floats with the same shape as `segments`.
    tube_points :   int
                    Number of points making up the circle of the cross-section
                    of the tube.
    use_normals :   bool
                    If True will rotate tube along it's curvature.

    Returns
    -------
    vertices :      np.ndarray
    faces :         np.ndarray

    """
    vertices = np.empty((0, 3), dtype=np.float64)
    indices = np.empty((0, 3), dtype=np.uint32)

    if not isinstance(radii, Iterable):
        radii = [[radii] * len(points) for points in segments]

    for points, radius in zip(segments, radii):
        # Need to make sure points are floats
        points = np.array(points).astype(float)

        # Skip single points
        if len(points) < 2:
            continue

        if use_normals:
            _, normals, binormals = _frenet_frames(points)
        else:
            _ = normals = binormals = np.ones((len(points), 3))

        n_segments = len(points) - 1

        if not isinstance(radius, Iterable):
            radius = [radius] * len(points)

        radius = np.array(radius)

        # Vertices for each point on the circle
        verts = np.repeat(points, tube_points, axis=0)

        v = np.arange(tube_points, dtype=np.float64) / tube_points * 2 * np.pi

        all_cx = (
            radius
            * -1.0
            * np.tile(np.cos(v), points.shape[0]).reshape(
                (tube_points, points.shape[0]), order="F"
            )
        ).T
        cx_norm = (all_cx[:, :, np.newaxis] * normals[:, np.newaxis, :]).reshape(
            verts.shape
        )

        all_cy = (
            radius
            * np.tile(np.sin(v), points.shape[0]).reshape(
                (tube_points, points.shape[0]), order="F"
            )
        ).T
        cy_norm = (all_cy[:, :, np.newaxis] * binormals[:, np.newaxis, :]).reshape(
            verts.shape
        )

        verts = verts + cx_norm + cy_norm

        # Generate indices for the first segment
        ix = np.arange(0, tube_points)

        # Repeat indices n_segments-times
        ix = np.tile(ix, n_segments)

        # Offset indices by number segments and tube points
        offsets = np.repeat((np.arange(0, n_segments)) * tube_points, tube_points)
        ix += offsets

        # Turn indices into faces
        ix_a = ix
        ix_b = ix + tube_points

        ix_c = ix_b.reshape((n_segments, tube_points))
        ix_c = np.append(ix_c[:, 1:], ix_c[:, [0]], axis=1)
        ix_c = ix_c.ravel()

        ix_d = ix_a.reshape((n_segments, tube_points))
        ix_d = np.append(ix_d[:, 1:], ix_d[:, [0]], axis=1)
        ix_d = ix_d.ravel()

        faces1 = np.concatenate((ix_a, ix_b, ix_d), axis=0).reshape(
            (n_segments * tube_points, 3), order="F"
        )
        faces2 = np.concatenate((ix_b, ix_c, ix_d), axis=0).reshape(
            (n_segments * tube_points, 3), order="F"
        )

        faces = np.append(faces1, faces2, axis=0)

        # Offset faces against already existing vertices
        faces += vertices.shape[0]

        # Add vertices and faces to total collection
        vertices = np.append(vertices, verts, axis=0)
        indices = np.append(indices, faces, axis=0)

    return vertices, indices


def _frenet_frames(points):
    """Calculate and return the tangents, normals and binormals for the tube.

    This code was modified from the vispy library.

    """
    tangents = np.zeros((len(points), 3))
    normals = np.zeros((len(points), 3))

    epsilon = 0.0001

    # Compute tangent vectors for each segment
    tangents = np.roll(points, -1, axis=0) - np.roll(points, 1, axis=0)

    tangents[0] = points[1] - points[0]
    tangents[-1] = points[-1] - points[-2]

    mags = np.sqrt(np.sum(tangents * tangents, axis=1))

    # Make sure we don't have any zeros in `mags` that would mess with the
    # subdivision ->  we will set those to the smallest possible value
    mags[mags == 0] = np.finfo(mags.dtype).resolution

    tangents /= mags[:, np.newaxis]

    # Get initial normal and binormal
    t = np.abs(tangents[0])

    smallest = np.argmin(t)
    normal = np.zeros(3)
    normal[smallest] = 1.0

    vec = np.cross(tangents[0], normal)
    normals[0] = np.cross(tangents[0], vec)

    all_vec = np.cross(tangents[:-1], tangents[1:])
    all_vec_norm = np.linalg.norm(all_vec, axis=1)

    # Normalise vectors if necessary
    where = all_vec_norm > epsilon
    all_vec[where, :] /= all_vec_norm[where].reshape((sum(where), 1))

    # Precompute inner dot product
    dp = np.sum(tangents[:-1] * tangents[1:], axis=1)
    # Clip
    cl = np.clip(dp, -1, 1)
    # Get theta
    th = np.arccos(cl)

    # Compute normal and binormal vectors along the path
    for i in range(1, len(points)):
        normals[i] = normals[i - 1]

        vec_norm = all_vec_norm[i - 1]
        vec = all_vec[i - 1]
        if vec_norm > epsilon:
            normals[i] = rotate(-np.degrees(th[i - 1]), vec)[:3, :3].dot(normals[i])

    binormals = np.cross(tangents, normals)

    return tangents, normals, binormals


def rotate(angle, axis, dtype=None):
    """Generate a 4x4 rotation matrix for rotation about a vector.

    Modified from `vispy.utils.transforms`.

    Parameters
    ----------
    angle :     float
                The angle of rotation, in degrees.
    axis :      ndarray
                The x, y, z coordinates of the axis direction vector.

    Returns
    -------
    M :     ndarray
            Transformation matrix describing the rotation.

    """
    angle = np.radians(angle)
    assert len(axis) == 3
    x, y, z = axis / np.linalg.norm(axis)
    c, s = math.cos(angle), math.sin(angle)
    cx, cy, cz = (1 - c) * x, (1 - c) * y, (1 - c) * z
    M = np.array(
        [
            [cx * x + c, cy * x - z * s, cz * x + y * s, 0.0],
            [cx * y + z * s, cy * y + c, cz * y - x * s, 0.0],
            [cx * z - y * s, cy * z + x * s, cz * z + c, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ],
        dtype,
    ).T
    return M



from typing import List, Tuple, Dict

import pytest
import pandas as pd
import numpy as np

import navis
from navis.connectivity import NeuronConnector
from navis import NeuronList


def test_neuron_connector():
    nrns = []
    for n in navis.example_neurons():
        n.name = f"{n.name}_{n.id}"
        nrns.append(n)

    conn = NeuronConnector(nrns)

    adj = conn.to_adjacency()
    assert len(adj) == len(nrns) + 1
    assert adj.to_numpy().sum() == 0

    dg = conn.to_digraph()
    assert dg.number_of_nodes() == len(nrns) + 1
    assert dg.number_of_edges() == 0

    mdg = conn.to_multidigraph()
    assert mdg.number_of_nodes() == dg.number_of_nodes()
    assert mdg.number_of_edges() == 0


def path_neuron(path: List[int]):
    nrn = navis.TreeNeuron(None)
    nrn.name = "".join(str(n) for n in path)
    dtypes = {
        "node_id": np.uint64,
        "parent_id": np.int64,
        "x": float,
        "y": float,
        "z": float,
    }
    prev = -1
    rows = []
    for n in path:
        rows.append([n, prev, 0, 0, 0])
        prev = n
    df = pd.DataFrame(rows, columns=list(dtypes)).astype(dtypes)
    nrn.nodes = df
    return nrn


def add_connectors(
    nrn: navis.TreeNeuron,
    incoming: List[Tuple[int, int]],
    outgoing: List[Tuple[int, int]],
):
    """Add connectors to a neuron.

    Parameters
    ----------
    incoming : list[tuple[int, int]]
        List of connector_id, node_id pairs.
    outgoing : list[tuple[int, int]]
        List of connector_id, node_id pairs
    """
    dtypes = {
        "connector_id": np.uint64,
        "node_id": np.uint64,
        "x": float,
        "y": float,
        "z": float,
        "type": np.uint64,
    }
    rows = []
    for tc_rel, ids in [(1, incoming), (0, outgoing)]:
        for conn, node in ids:
            rows.append([conn, node, 0, 0, 0, tc_rel])
    df = pd.DataFrame(rows, columns=list(dtypes)).astype(dtypes)
    nrn.connectors = df


@pytest.fixture
def simple_network():
    """
    2 neurons, "456" and "789".
    4 connectors, 0/1/2/3.

        4     7
        |     |
    =0> 5 =1> 8 =2>
        |     |
        6 =3> 9
    """
    n456 = path_neuron([4, 5, 6])
    add_connectors(n456, [(0, 5)], [(1, 5), (3, 6)])
    n789 = path_neuron([7, 8, 9])
    add_connectors(n789, [(1, 8), (3, 9)], [(2, 8)])
    return [n456, n789]


def test_neuron_connector_synthetic(simple_network):
    nconn = NeuronConnector(simple_network)

    adj = nconn.to_adjacency()
    assert len(adj) == len(simple_network) + 1
    assert adj.to_numpy().sum() == 4

    expected_edges = sorted([
        ("__OTHER__", "456"),
        ("456", "789"),
        ("456", "789"),
        ("789", "__OTHER__"),
    ])

    dg = nconn.to_digraph()
    assert dg.number_of_nodes() == len(simple_network) + 1
    assert dg.number_of_edges() == 3
    assert set(dg.edges()) == set(expected_edges)

    mdg = nconn.to_multidigraph()
    assert mdg.number_of_nodes() == dg.number_of_nodes()
    assert mdg.number_of_edges() == 4
    assert sorted(mdg.edges()) == expected_edges


def test_neuron_connector_real(
    neuron_connections: Tuple[NeuronList, Dict[int, Dict[int, int]]]
):
    nl, exp = neuron_connections
    nc = NeuronConnector(nl)
    dg = nc.to_digraph(include_other=False)
    for pre_n, post_n, edata in dg.edges(data=True):
        pre_skid = dg.nodes[pre_n]["neuron"].id
        post_skid = dg.nodes[post_n]["neuron"].id
        n = edata["weight"]

        assert exp[pre_skid][post_skid] == n



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
import pandas as pd
import numpy as np

from pathlib import Path
from typing import List, Union, Optional

from .. import config, core

__all__ = ["read_parquet", "write_parquet", "scan_parquet"]

# Set up logging
logger = config.get_logger(__name__)

SKELETON_COLUMNS = ('node_id', 'x', 'y', 'z', 'radius', 'parent_id', 'neuron')
NA_VALUES = (None, 'None')
META_DATA = ('name', 'units', 'soma')  # meta data to write for each neuron

INT_TYPES = (int, np.int8, np.int16, np.int32, np.int64)


def scan_parquet(file: Union[str, Path]):
    """Scan parquet file.

    Parameters
    ----------
    file :              str
                        File to be scan.

    Returns
    -------
    pd.DataFrame
                        Summary of file's content.

    See Also
    --------
    [`navis.write_parquet`][]
                        Export neurons as parquet files.
    [`navis.read_parquet`][]
                        Read parquet file into neurons.

    Examples
    --------
    See [`navis.write_parquet`][] for examples.

    """
    try:
        import pyarrow.parquet as pq
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            'Reading parquet files requires the pyarrow library:\n'
            ' pip3 install pyarrow')

    f = Path(file).expanduser()
    if not f.is_file():
        raise FileNotFoundError(f'File "{f}" does not exist.')

    metadata = pq.read_metadata(f)

    try:
        meta = {k.decode(): v.decode() for k, v in metadata.metadata.items()}
    except BaseException:
        logger.warning(f'Unable to decode meta data for parquet file {f}')

    # Parse meta data
    ids = [v for k, v in meta.items() if k.endswith(':id') and not k.startswith('_')]
    records = {i: {} for i in ids}
    for k, v in meta.items():
        if k.startswith('_'):
            continue
        if ':' not in k:
            continue

        id, prop = k.split(':')

        if id not in records:  # there might be an "ARROW:schema" entry
            continue

        records[id][prop] = v

    # Turn into DataFrame
    df =  pd.DataFrame.from_records(list(records.values()))

    # Move ID column to front
    ids = df['id']
    df.drop(labels=['id'], axis=1, inplace=True)
    df.insert(0, 'id', ids)

    return df


def read_parquet(f: Union[str, Path],
                 read_meta: bool = True,
                 limit: Optional[int] = None,
                 subset: Optional[List[Union[str, int]]] = None,
                 progress=True
                 ) -> 'core.NeuronObject':
    """Read parquet file into Neuron/List.

    See [here](https://github.com/navis-org/navis/blob/master/navis/io/pq_io.md)
    for format specifications.

    Parameters
    ----------
    f :                 str
                        File to be read.
    read_meta :         bool
                        Whether to read neuron meta data stored in the parquet
                        file (e.g. name or units). Defaults to True but can be
                        switched off in case there are any issues.
    limit :             int, optional
                        If reading from a file containing multiple neurons you
                        can use this parameter to read only the first `limit`
                        neurons. Useful if wanting to get a sample from a large
                        library of neurons.
    subset :            str | int | list thereof
                        If the parquet file contains multiple neurons you can
                        use this to select the IDs of the neurons to load. Only
                        works if the parquet file actually contains multiple
                        neurons.

    Returns
    -------
    navis.TreeNeuron/Dotprops
                        If parquet file contains a single neuron.
    navis.NeuronList
                        If parquet file contains multiple neurons.

    See Also
    --------
    [`navis.write_parquet`][]
                        Export neurons as parquet files.
    [`navis.scan_parquet`][]
                        Scan parquet file for its contents.

    Examples
    --------
    See [`navis.write_parquet`][] for examples.

    """
    f = Path(f).expanduser()
    if not f.is_file():
        raise FileNotFoundError(f'File "{f}" does not exist.')

    try:
        import pyarrow.parquet as pq
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            'Reading parquet files requires the pyarrow library:\n'
            ' pip3 install pyarrow')

    if limit is not None:
        if subset not in (None, False):
            raise ValueError('You can provide either a `subset` or a `limit` but '
                             'not both.')
        scan = scan_parquet(f)
        subset = scan.id.values[:limit]

    if isinstance(subset, (pd.Series)):
        subset = subset.values

    # Read the table
    if subset is None or subset is False:
        table = pq.read_table(f)
    elif isinstance(subset, (str, int)):
        table = pq.read_table(f, filters=[("neuron", "=", subset)])
    elif isinstance(subset, (list, np.ndarray)):
        table = pq.read_table(f, filters=[("neuron", "in", subset)])
    else:
        raise TypeError(f'`subset` must be int, str or iterable, got "{type(subset)}')

    # Extract meta data (will be byte encoded)
    if read_meta:
        metadata = {k.decode(): v.decode() for k, v in table.schema.metadata.items()}
    else:
        metadata = {}

    # Extract neuron meta data once here instead of for every neuron individually
    # Meta data is encoded as {"{ID}_{PROPERTY}": VALUE}
    # Here we pre-emptively turn this into {(ID, PROPERTY): VALUE}
    # Note that we're dropping "private" properties where the key starts with "_"
    neuron_meta = {tuple(k.split(':')): v for k, v in metadata.items() if not k.startswith('_')}

    # Convert to pandas
    table = table.to_pandas()

    # Check if we're doing skeletons or dotprops
    if 'node_id' in table.columns:
        _extract_neuron = _extract_skeleton
    elif 'vect_x' in table.columns:
        _extract_neuron = _extract_dotprops
    else:
        raise TypeError('Unable to extract neuron from parquet file with '
                        f'columns {table.columns}')

    # If this is a single neuron
    if 'neuron' not in table.columns:
        if metadata:
            id = [v for k, v in metadata.items() if k[1] == 'id'][0]
        else:
            id = '0'  # <-- generic ID as fallback if we don't have metadata
        return _extract_neuron(table, id, neuron_meta)
    else:
        neurons = []
        # Note: this could be done in threads
        for i, (id, this_table) in enumerate(config.tqdm(table.groupby('neuron'),
                                             disable=not progress,
                                             leave=False,
                                             desc='Making nrn')):
            this_table = this_table.drop("neuron", axis=1)
            neurons.append(_extract_neuron(this_table, id, neuron_meta))
        return core.NeuronList(neurons)


def _extract_skeleton(nodes, id, metadata):
    """Extract a single skeleton."""
    # Meta data is encoded as "{ID}_{PROPERTY}"
    str_id = str(id)
    this_meta = {k[1]: v for k, v in metadata.items() if k[0] == str_id}
    # Drop "Nones"
    this_meta = {k: v for k, v in this_meta.items() if v != "None"}

    # The soma needs to be added separately because it is typically stored as
    # list (e.g. [0]) which the TreeNeuron initialisation doesn't like
    if "soma" in this_meta:
        soma = this_meta.pop("soma")
        # Parse a list string (e.g. "[1]") back into a list
        if soma.startswith('['):
            soma = [_try_int(i.strip()) for i in soma[1:-1].split(',')]
        else:
            soma = _try_int(soma)
    else:
        soma = None

    # Make the neuron
    this_meta['id'] = id
    tn = core.TreeNeuron(nodes, **this_meta)

    # Fix soma
    if soma:
        tn.soma = soma
    else:
        tn.soma = None

    return tn


def _extract_dotprops(table, id, metadata):
    """Extract a single dotprop."""
    # Meta data is encoded as "{ID}_{PROPERTY}"
    str_id = str(id)
    this_meta = {k[1]: v for k, v in metadata.items() if k[0] == str_id}
    # Drop "Nones"
    this_meta = {k: v for k, v in this_meta.items() if v != "None"}

    # Make the neuron
    this_meta['id'] = id
    this_meta['k'] = this_meta.get('k', 5)  # <- set a default K of 5

    if 'vect_x' in table:
        this_meta['vect'] = table[['vect_x', 'vect_y', 'vect_z']].values
    if 'alpha' in table:
        this_meta['alpha'] = table['alpha'].values

    return core.Dotprops(table[['x', 'y', 'z']].values,
                       **this_meta)


def _try_int(x):
    """Try converting `x` into an integer."""
    try:
        return int(x)
    except ValueError:
        return x


def _int_to_bytes(x, bits=64):
    """Convert integer to bytes."""
    return int(x).to_bytes(bits, 'big')


def _bytes_to_int(x):
    """Convert bytes to integer."""
    return int.from_bytes(x, "big")


def write_parquet(x: 'core.NeuronObject',
                  filepath: Union[str, Path],
                  write_meta: bool = True) -> None:
    """Write TreeNeuron(s) or Dotprops to parquet file.

    See [here](https://github.com/navis-org/navis/blob/master/navis/io/pq_io.md)
    for format specifications.

    Parameters
    ----------
    x :                 TreeNeuron | Dotprop | NeuronList thereof
                        Neuron(s) to save. If NeuronList must contain either
                        only TreeNeurons or only Dotprops.
    filepath :          str | pathlib.Path
                        Destination for the file.
    write_meta :        bool | list of str
                        Whether to also write neuron properties to file. By
                        default this is `.name`, `.units` and `.soma`. You can
                        change which properties are written by providing them as
                        list of strings.

    See Also
    --------
    [`navis.read_parquet`][]
                        Import skeleton from parquet file.
    [`navis.scan_parquet`][]
                        Scan parquet file for its contents.

    Examples
    --------
    Save a bunch of skeletons:

    >>> import navis
    >>> nl = navis.example_neurons(3, kind='skeleton')
    >>> navis.write_parquet(nl, tmp_dir / 'skeletons.parquet')

    Inspect that file's content

    >>> import navis
    >>> contents = navis.scan_parquet(tmp_dir / 'skeletons.parquet')
    >>> contents                                                # doctest: +SKIP
               id        units       name    soma
    0   722817260  8 nanometer  DA1_lPN_R     NaN
    1  1734350908  8 nanometer  DA1_lPN_R     [6]
    2  1734350788  8 nanometer  DA1_lPN_R  [4177]

    Read the skeletons back in

    >>> import navis
    >>> nl = navis.read_parquet(tmp_dir / 'skeletons.parquet')
    >>> len(nl)
    3

    """
    filepath = Path(filepath).expanduser()

    # Make sure inputs are only TreeNeurons or Dotprops
    if isinstance(x, core.NeuronList):
        types = x.types
        if types == (core.TreeNeuron,):
            _write_parquet = _write_parquet_skeletons
        elif types == (core.Dotprops, ):
            _write_parquet = _write_parquet_dotprops
        else:
            raise TypeError('Can only write either TreeNeurons or Dotprops to '
                            f'parquet but NeuronList contains {types}')
        if x.is_degenerated:
            raise ValueError('NeuronList must not contain non-unique IDs')
    else:
        if isinstance(x, (core.TreeNeuron, )):
            _write_parquet = _write_parquet_skeletons
        elif isinstance(x, (core.Dotprops, )):
            _write_parquet = _write_parquet_dotprops
        else:
            raise TypeError('Can only write TreeNeurons or Dotprops to parquet, '
                            f'got "{type(x)}"')

    return _write_parquet(x, filepath=filepath, write_meta=write_meta)


def _write_parquet_skeletons(x: 'core.TreeNeuron',
                             filepath: Union[str, Path],
                             write_meta: bool = True,
                             ) -> None:
    """Write TreeNeurons to parquet file."""
    try:
        import pyarrow as pa
        import pyarrow.parquet as pq
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            'Writing parquet files requires the pyarrow library:\n'
            ' pip3 install pyarrow'
            )

    # Generate node table
    nodes = x.nodes[x.nodes.columns[np.isin(x.nodes.columns, SKELETON_COLUMNS)]]

    # Convert to pyarrow table
    table = pa.Table.from_pandas(nodes)

    # Compile metadata
    metadata = _compile_meta(x, write_meta=write_meta)

    # Generate a schema with the new meta data
    schema = pa.schema([table.schema.field(i) for i in range(len(table.schema))],
                       metadata=metadata)

    return pq.write_table(table.cast(schema), filepath)


def _write_parquet_dotprops(x: 'core.Dotprops',
                            filepath: Union[str, Path],
                            write_meta: bool = True,
                            ) -> None:
    """Write Dotprops to parquet file.

    Examples
    --------
    We will test writing dotprops here instead of the main function

    >>> import navis
    >>> nl = navis.example_neurons(3, kind='skeleton')
    >>> dp = navis.make_dotprops(nl, k=5)
    >>> navis.write_parquet(dp, tmp_dir / 'dotprops.parquet')
    >>> dp2 = navis.read_parquet(tmp_dir / 'dotprops.parquet')
    >>> assert len(dp) == len(dp2)
    >>> assert all([i in dp2.id for i in dp.id])

    """
    try:
        import pyarrow as pa
        import pyarrow.parquet as pq
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            'Writing parquet files requires the pyarrow library:\n'
            ' pip3 install pyarrow'
            )

    # Make sure we're working with a list, not a single neuron
    x = core.NeuronList(x)

    # Generate table
    table = pd.DataFrame(np.vstack(x.points), columns=['x', 'y', 'z'])

    if all(x.has_vect):
        table = pd.concat((table,
                           pd.DataFrame(np.vstack(x.vect),
                                        columns=['vect_x', 'vect_y', 'vect_z'])
                                        ),
                          axis=1)

    if all(x.has_alpha):
        table['alpha'] = np.concatenate(x.alpha)

    # Add neuron ID
    table['neuron'] = np.repeat(x.id, x.n_points)

    # Convert to pyarrow table
    table = pa.Table.from_pandas(table)

    # Compile metadata
    metadata = _compile_meta(x, write_meta=write_meta)

    # Generate a schema with the new meta data
    schema = pa.schema([table.schema.field(i) for i in range(len(table.schema))],
                       metadata=metadata)

    return pq.write_table(table.cast(schema), filepath)


def _compile_meta(x: Union['core.BaseNeuron', 'core.NeuronList'],
                  write_meta: bool
                  ) -> dict:
    """Compile meta data for writing to parquet file."""
    metadata = {}
    for n in core.NeuronList(x):
        # ID is always written to file and it has to be a string
        if isinstance(n.id, INT_TYPES):
            metadata[f'{n.id}:id'] = str(n.id) #_int_to_bytes(n.id)
        else:
            metadata[f'{n.id}:id'] = str(n.id)

        # If not write_meta, only ID is written to file
        if not write_meta:
            continue

        if isinstance(write_meta, (list, np.ndarray, tuple)):
            attrs = write_meta
        else:
            attrs = META_DATA

        for p in attrs:
            if not getattr(n, p, None):
                continue
            # We're mapping meta data as "{ID}_{property}"
            # e.g. {"1734350788_name": "DA1_lPN_R"}
            metadata[f'{n.id}:{p}'] = str(getattr(n, p, None))

    return metadata


#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from .json_io import read_json, write_json
from .swc_io import read_swc, write_swc
from .nrrd_io import read_nrrd, write_nrrd
from .precomputed_io import read_precomputed, write_precomputed
from .hdf_io import read_h5, write_h5, inspect_h5
from .rda_io import read_rda
from .nmx_io import read_nmx, read_nml
from .mesh_io import read_mesh, write_mesh
from .tiff_io import read_tiff
from .pq_io import read_parquet, write_parquet, scan_parquet

__all__ = ['read_json', 'write_json',
           'read_swc', 'write_swc',
           'read_nrrd', 'write_nrrd',
           'read_h5', 'write_h5', 'inspect_h5',
           'read_precomputed', 'write_precomputed',
           'read_tiff',
           'read_rda',
           'read_nmx', 'read_nml',
           'read_mesh', 'write_mesh',
           'read_parquet', 'write_parquet', 'scan_parquet']



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import io

import numpy as np

from typing import Union, Iterable, Optional, Dict, Any
from typing_extensions import Literal
from urllib3 import HTTPResponse

from .. import config, utils, core
from . import base

# Set up logging
logger = config.logger

DEFAULT_FMT = "{name}.tif"


class TiffReader(base.ImageReader):
    def __init__(
        self,
        output: Literal["voxels", "dotprops", "raw"] = "voxels",
        channel: int = 0,
        threshold: Optional[Union[int, float]] = None,
        thin: bool = False,
        dotprop_kwargs: Dict[str, Any] = {},
        fmt: str = DEFAULT_FMT,
        errors: str = "raise",
        attrs: Optional[Dict[str, Any]] = None,
    ):
        if not fmt.endswith(".tif") and not fmt.endswith(".tiff"):
            raise ValueError('`fmt` must end with ".tif" or ".tiff"')

        super().__init__(
            fmt=fmt,
            attrs=attrs,
            file_ext=(".tif", ".tiff"),
            name_fallback="TIFF",
            read_binary=True,
            output=output,
            threshold=threshold,
            thin=thin,
            dotprop_kwargs=dotprop_kwargs,
            errors=errors,
        )
        self.channel = channel

    def format_output(self, x):
        # This function replaces the BaseReader.format_output()
        # This is to avoid trying to convert multiple (image, header) to NeuronList
        if self.output == "raw":
            return x
        elif x:
            return core.NeuronList([n for n in x if n])
        else:
            return core.NeuronList([])

    @base.handle_errors
    def read_buffer(
        self, f, attrs: Optional[Dict[str, Any]] = None
    ) -> Union[np.ndarray, "core.Dotprops", "core.VoxelNeuron"]:
        """Read buffer into (image, header) or a neuron.

        Parameters
        ----------
        f :         IO
                    Readable buffer (must be bytes).
        attrs :     dict | None
                    Arbitrary attributes to include in the neuron.

        Returns
        -------
        core.Dotprops | core.VoxelNeuron | np.ndarray

        """
        import tifffile

        if isinstance(f, HTTPResponse):
            f = io.StringIO(f.content)

        if isinstance(f, bytes):
            f = io.BytesIO(f)

        with tifffile.TiffFile(f) as tif:
            # The header contains some but not all the info
            if hasattr(tif, "imagej_metadata") and tif.imagej_metadata is not None:
                header = tif.imagej_metadata
            else:
                header = {}

            # Read the x/y resolution from the first "page" (i.e. the first slice)
            res = tif.pages[0].resolution
            # Resolution to spacing
            header["xy_spacing"] = (1 / res[0], 1 / res[1])

            # Get the axes; this will be something like "ZCYX" where:
            # Z = slices, C = channels, Y = rows, X = columns, S = color(?), Q = empty(?)
            axes = tif.series[0].axes

            # Generate volume
            data = tif.asarray()

        if self.output == "raw":
            return data, header

        # Drop "Q" axes if they have dimenions of 1 (we're assuming these are empty)
        while "Q" in axes and data.shape[axes.index("Q")] == 1:
            data = np.squeeze(data, axis=axes.index("Q"))
            axes = axes.replace("Q", "", 1)  # Only remove the first occurrence
        if "C" in axes:
            # Extract the requested channel from the volume
            data = data.take(self.channel, axis=axes.index("C"))
            axes = axes.replace("C", "")

        # At this point we expect 3D data
        if data.ndim != 3:
            raise ValueError(f'Expected 3D greyscale data, got {data.ndim} ("{axes}").')

        # Swap axes to XYZ order
        order = []
        for a in ("X", "Y", "Z"):
            if a not in axes:
                logger.warning(
                    f'Expected axes to contain "Z", "Y", and "X", got "{axes}". '
                    "Axes will not be automatically reordered."
                )
                order = None
                break
            order.append(axes.index(a))
        if order:
            data = np.transpose(data, order)

        # Try parsing units - this is modelled after the tif files you get from ImageJ
        units = None
        space_units = None
        voxdim = np.array([1, 1, 1], dtype=np.float64)
        if "spacing" in header:
            voxdim[2] = header["spacing"]
        if "xy_spacing" in header:
            voxdim[:2] = header["xy_spacing"]
        if "unit" in header:
            space_units = header["unit"]
            units = [f"{m} {space_units}" for m in voxdim]
        else:
            units = voxdim

        return self.convert_image(data, attrs, header, voxdim, units, space_units)


def read_tiff(
    f: Union[str, Iterable],
    output: Union[Literal["voxels"], Literal["dotprops"], Literal["raw"]] = "voxels",
    channel: int = 0,
    threshold: Optional[Union[int, float]] = None,
    thin: bool = False,
    include_subdirs: bool = False,
    parallel: Union[bool, int] = "auto",
    fmt: str = "{name}.tif",
    limit: Optional[int] = None,
    errors: str = "raise",
    **dotprops_kwargs,
) -> "core.NeuronObject":
    """Create Neuron/List from TIFF file.

    Requires `tifffile` library which is not automatically installed!

    Parameters
    ----------
    f :                 str | iterable
                        Filename(s) or folder. If folder, will import all
                        `.tif` files.
    output :            "voxels" | "dotprops" | "raw"
                        Determines function's output. See Returns for details.
    channel :           int
                        Which channel to import. Ignored if file has only one
                        channel or when `output="raw". Can use e.g. -1 to
                        get the last channel.
    threshold :         int | float | None
                        For `output='dotprops'` only: a threshold to filter
                        low intensity voxels. If `None`, no threshold is
                        applied and all values > 0 are converted to points.
    thin :              bool
                        For `output='dotprops'` only: if True, will thin the
                        point cloud using `skimage.morphology.skeletonize`
                        after thresholding. Requires `scikit-image`.
    include_subdirs :   bool, optional
                        If True and `f` is a folder, will also search
                        subdirectories for `.tif` files.
    parallel :          "auto" | bool | int,
                        Defaults to `auto` which means only use parallel
                        processing if more than 10 TIFF files are imported.
                        Spawning and joining processes causes overhead and is
                        considerably slower for imports of small numbers of
                        neurons. Integer will be interpreted as the number of
                        cores (otherwise defaults to `os.cpu_count() - 2`).
    fmt :               str
                        Formatter to specify how filenames are parsed into neuron
                        attributes. Some illustrative examples:
                          - `{name}` (default) uses the filename
                            (minus the suffix) as the neuron's name property
                          - `{id}` (default) uses the filename as the neuron's ID
                            property
                          - `{name,id}` uses the filename as the neuron's
                            name and ID properties
                          - `{name}.{id}` splits the filename at a "."
                            and uses the first part as name and the second as ID
                          - `{name,id:int}` same as above but converts
                            into integer for the ID
                          - `{name}_{myproperty}` splits the filename at
                            "_" and uses the first part as name and as a
                            generic "myproperty" property
                          - `{name}_{}_{id}` splits the filename at
                            "_" and uses the first part as name and the last as
                            ID. The middle part is ignored.

                        Throws a ValueError if pattern can't be found in
                        filename.
    limit :             int | str | slice | list, optional
                        When reading from a folder or archive you can use this parameter to
                        restrict the which files read:
                         - if an integer, will read only the first `limit` NMX files
                           (useful to get a sample from a large library of skeletons)
                         - if a string, will interpret it as filename (regex) pattern
                           and only read files that match the pattern; e.g. `limit='.*_R.*'`
                           will only read files that contain `_R` in their filename
                         - if a slice (e.g. `slice(10, 20)`) will read only the files in
                           that range
                         - a list is expected to be a list of filenames to read from
                           the folder/archive
    errors :            "raise" | "log" | "ignore"
                        If "log" or "ignore", errors will not be raised and the
                        mesh will be skipped. Can result in empty output.

    **dotprops_kwargs
                        Keyword arguments passed to [`navis.make_dotprops`][]
                        if `output='dotprops'`. Use this to adjust e.g. the
                        number of nearest neighbors used for calculating the
                        tangent vector by passing e.g. `k=5`.

    Returns
    -------
    navis.VoxelNeuron
                        If `output="voxels"` (default): requires TIFF data to
                        be 3-dimensional voxels. VoxelNeuron will have TIFF file
                        info as `.tiff_header` attribute.
    navis.Dotprops
                        If `output="dotprops"`. Dotprops will contain TIFF
                        header as `.tiff_header` attribute.
    navis.NeuronList
                        If import of multiple TIFF will return NeuronList of
                        Dotprops/VoxelNeurons.
    (image, header)     (np.ndarray, OrderedDict)
                        If `output='raw'` return raw data contained in TIFF
                        file.

    """
    try:
        import tifffile
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "`navis.read_tiff` requires the `tifffile` library:\n"
            "  pip3 install tifffile -U"
        )

    utils.eval_param(
        output, name="output", allowed_values=("raw", "dotprops", "voxels")
    )

    if parallel == "auto":
        # Set a lower threshold of 10 on parallel processing for TIFFs (default is 200)
        parallel = ("auto", 10)

    reader = TiffReader(
        channel=channel,
        output=output,
        threshold=threshold,
        thin=thin,
        fmt=fmt,
        dotprop_kwargs=dotprops_kwargs,
        errors=errors,
    )
    return reader.read_any(f, include_subdirs, parallel, limit=limit)
